import pydantic
import pytest

from ..adapters.mapping import MapAdapter
from ..client import from_tree
from ..structures.core import Spec


def test_specs():
    tree = MapAdapter({}, specs=[Spec("spec_test")])
    c = from_tree(tree)
    assert c.specs == [Spec("spec_test")]


def test_specs_give_as_str():
    tree = MapAdapter({}, specs=["spec_test"])
    c = from_tree(tree)
    assert c.specs == [Spec("spec_test")]


def test_specs_with_version():
    tree = MapAdapter({}, specs=[Spec("spec_test", version="1.1")])
    c = from_tree(tree)
    assert c.specs == [Spec("spec_test", version="1.1")]


def test_references():
    tree = MapAdapter(
        {}, references=[{"label": "ref_test", "url": "https://example.com"}]
    )
    c = from_tree(tree)
    assert c.references == [{"label": "ref_test", "url": "https://example.com"}]


def test_bad_reference():
    # an invalid URL
    tree = MapAdapter({}, references=[{"label": "ref_test", "url": "not a URL"}])
    with pytest.raises(pydantic.error_wrappers.ValidationError):
        from_tree(tree)

    # a dict instead of a list of dicts
    tree = MapAdapter(
        {}, references={"label": "ref_test", "url": "https://example.com"}
    )
    with pytest.raises(pydantic.error_wrappers.ValidationError):
        from_tree(tree)

    # dict has too many items
    tree = MapAdapter(
        {},
        references=[
            {"label": "ref_test", "url": "https://example.com", "extra": "oops"}
        ],
    )
    with pytest.raises(pydantic.error_wrappers.ValidationError):
        from_tree(tree)

    # dict has not enough items
    tree = MapAdapter({}, references=[{}])
    with pytest.raises(pydantic.error_wrappers.ValidationError):
        from_tree(tree)



"""
This tests tiled's writing routes with an in-memory store.

Persistent stores are being developed externally to the tiled package.
"""

import dask.dataframe
import numpy
import pandas.testing
import sparse

from ..client import from_tree, record_history
from ..queries import Key
from ..structures.core import Spec
from ..structures.sparse import COOStructure
from ..validation_registration import ValidationRegistry
from .utils import fail_with_status_code
from .writable_adapters import WritableMapAdapter

API_KEY = "secret"
validation_registry = ValidationRegistry()
validation_registry.register("SomeSpec", lambda *args, **kwargs: None)


def test_write_array_full():

    tree = WritableMapAdapter({})
    client = from_tree(
        tree,
        api_key=API_KEY,
        authentication={"single_user_api_key": API_KEY},
        validation_registry=validation_registry,
    )

    a = numpy.ones((5, 7))

    metadata = {"scan_id": 1, "method": "A"}
    specs = [Spec("SomeSpec")]
    references = [{"label": "test", "url": "http://www.test.com"}]
    with record_history() as history:
        client.write_array(a, metadata=metadata, specs=specs, references=references)
    # one request for metadata, one for data
    assert len(history.requests) == 1 + 1

    results = client.search(Key("scan_id") == 1)
    result = results.values().first()
    result_array = result.read()

    numpy.testing.assert_equal(result_array, a)
    assert result.metadata == metadata
    assert result.specs == specs
    assert result.references == references


def test_write_large_array_full():
    "Test that a large array is chunked"

    tree = WritableMapAdapter({})
    client = from_tree(
        tree,
        api_key=API_KEY,
        authentication={"single_user_api_key": API_KEY},
        validation_registry=validation_registry,
    )

    a = numpy.ones(100, dtype=numpy.uint8)
    # Low the limit so we can test on small data, for speed.
    original = client._SUGGESTED_MAX_UPLOAD_SIZE
    client._SUGGESTED_MAX_UPLOAD_SIZE = a.nbytes - 1
    try:
        metadata = {"scan_id": 1, "method": "A"}
        specs = [Spec("SomeSpec")]
        references = [{"label": "test", "url": "http://www.test.com"}]
        with record_history() as history:
            client.write_array(a, metadata=metadata, specs=specs, references=references)
        # one request for metadata, more than one for data
        assert len(history.requests) > 1 + 1

        results = client.search(Key("scan_id") == 1)
        result = results.values().first()
        result_array = result.read()

        numpy.testing.assert_equal(result_array, a)
        assert result.metadata == metadata
        assert result.specs == specs
        assert result.references == references
    finally:
        client._SUGGESTED_MAX_UPLOAD_SIZE = original


def test_write_array_chunked():

    tree = WritableMapAdapter({})
    client = from_tree(
        tree,
        api_key=API_KEY,
        authentication={"single_user_api_key": API_KEY},
        validation_registry=validation_registry,
    )

    a = dask.array.arange(1500).reshape((50, 30)).rechunk((20, 15))

    metadata = {"scan_id": 1, "method": "A"}
    specs = [Spec("SomeSpec")]
    references = [{"label": "test", "url": "http://www.test.com"}]
    with record_history() as history:
        client.write_array(a, metadata=metadata, specs=specs, references=references)
    # one request for metadata, multiple for data
    assert len(history.requests) == 1 + a.npartitions

    results = client.search(Key("scan_id") == 1)
    result = results.values().first()
    result_array = result.read()

    numpy.testing.assert_equal(result_array, a.compute())
    assert result.metadata == metadata
    assert result.specs == specs
    assert result.references == references


def test_write_dataframe_full():

    tree = WritableMapAdapter({})
    client = from_tree(
        tree,
        api_key=API_KEY,
        authentication={"single_user_api_key": API_KEY},
        validation_registry=validation_registry,
    )

    data = {f"Column{i}": (1 + i) * numpy.ones(5) for i in range(5)}
    df = pandas.DataFrame(data)
    metadata = {"scan_id": 1, "method": "A"}
    specs = [Spec("SomeSpec")]
    references = [{"label": "test", "url": "http://www.test.com"}]

    with record_history() as history:
        client.write_dataframe(
            df, metadata=metadata, specs=specs, references=references
        )
    # one request for metadata, one for data
    assert len(history.requests) == 1 + 1

    results = client.search(Key("scan_id") == 1)
    result = results.values().first()
    result_dataframe = result.read()

    pandas.testing.assert_frame_equal(result_dataframe, df)
    assert result.metadata == metadata
    assert result.specs == specs
    assert result.references == references


def test_write_dataframe_partitioned():

    tree = WritableMapAdapter({})
    client = from_tree(
        tree,
        api_key=API_KEY,
        authentication={"single_user_api_key": API_KEY},
        validation_registry=validation_registry,
    )

    data = {f"Column{i}": (1 + i) * numpy.ones(10) for i in range(5)}
    df = pandas.DataFrame(data)
    ddf = dask.dataframe.from_pandas(df, npartitions=3)
    metadata = {"scan_id": 1, "method": "A"}
    specs = [Spec("SomeSpec")]
    references = [{"label": "test", "url": "http://www.test.com"}]

    with record_history() as history:
        client.write_dataframe(
            ddf, metadata=metadata, specs=specs, references=references
        )
    # one request for metadata, multiple for data
    assert len(history.requests) == 1 + 3

    results = client.search(Key("scan_id") == 1)
    result = results.values().first()
    result_dataframe = result.read()

    pandas.testing.assert_frame_equal(result_dataframe, df)
    assert result.metadata == metadata
    assert result.specs == specs
    assert result.references == references


def test_write_sparse_full():

    tree = WritableMapAdapter({})
    client = from_tree(
        tree,
        api_key=API_KEY,
        authentication={"single_user_api_key": API_KEY},
        validation_registry=validation_registry,
    )

    coo = sparse.COO(coords=[[0, 1], [2, 3]], data=[3.8, 4.0], shape=(4, 4))

    metadata = {"scan_id": 1, "method": "A"}
    specs = [Spec("SomeSpec")]
    references = [{"label": "test", "url": "http://www.test.com"}]
    with record_history() as history:
        client.write_sparse(
            coords=coo.coords,
            data=coo.data,
            shape=coo.shape,
            metadata=metadata,
            specs=specs,
            references=references,
        )
    # one request for metadata, one for data
    assert len(history.requests) == 1 + 1

    results = client.search(Key("scan_id") == 1)
    result = results.values().first()
    result_array = result.read()

    numpy.testing.assert_equal(result_array.todense(), coo.todense())
    assert result.metadata == metadata
    assert result.specs == specs
    assert result.references == references


def test_write_sparse_chunked():

    tree = WritableMapAdapter({})
    client = from_tree(
        tree,
        api_key=API_KEY,
        authentication={"single_user_api_key": API_KEY},
        validation_registry=validation_registry,
    )

    metadata = {"scan_id": 1, "method": "A"}
    specs = [Spec("SomeSpec")]
    references = [{"label": "test", "url": "http://www.test.com"}]
    N = 5
    with record_history() as history:
        x = client.new(
            "sparse",
            COOStructure(shape=(2 * N,), chunks=((N, N),)),
            metadata=metadata,
            specs=specs,
            references=references,
        )
        x.write_block(coords=[[2, 4]], data=[3.1, 2.8], block=(0,))
        x.write_block(coords=[[0, 1]], data=[6.7, 1.2], block=(1,))

    # one request for metadata, multiple for data
    assert len(history.requests) == 1 + 2

    results = client.search(Key("scan_id") == 1)
    result = results.values().first()
    result_array = result.read()
    assert numpy.array_equal(
        result_array.todense(),
        sparse.COO(
            coords=[[2, 4, N + 0, N + 1]], data=[3.1, 2.8, 6.7, 1.2], shape=(10,)
        ).todense(),
    )

    # numpy.testing.assert_equal(result_array, sparse.COO(coords=[0, 1, ]))
    assert result.metadata == metadata
    assert result.specs == specs
    assert result.references == references


def test_limits():
    "Test various limits on uploaded metadata."

    MAX_ALLOWED_SPECS = 20
    MAX_SPEC_CHARS = 255
    MAX_ALLOWED_REFERENCES = 20
    MAX_LABEL_CHARS = 255

    tree = WritableMapAdapter({})
    validation_registry = ValidationRegistry()
    for i in range(101):
        validation_registry.register(f"spec{i}", lambda *args, **kwargs: None)
    validation_registry.register("one_too_many", lambda *args, **kwargs: None)
    validation_registry.register("a" * MAX_SPEC_CHARS, lambda *args, **kwargs: None)
    validation_registry.register(
        "a" * (1 + MAX_SPEC_CHARS), lambda *args, **kwargs: None
    )
    client = from_tree(
        tree,
        api_key=API_KEY,
        authentication={"single_user_api_key": API_KEY},
        validation_registry=validation_registry,
    )

    # Up to 20 specs are allowed.
    max_allowed_specs = [f"spec{i}" for i in range(MAX_ALLOWED_SPECS)]
    x = client.write_array([1, 2, 3], specs=max_allowed_specs)
    x.update_metadata(specs=max_allowed_specs)  # no-op
    too_many_specs = max_allowed_specs + ["one_too_many"]
    with fail_with_status_code(422):
        client.write_array([1, 2, 3], specs=too_many_specs)
    with fail_with_status_code(422):
        x.update_metadata(specs=too_many_specs)

    # Specs cannot repeat.
    has_repeated_spec = ["spec0", "spec1", "spec0"]
    with fail_with_status_code(422):
        client.write_array([1, 2, 3], specs=has_repeated_spec)
    with fail_with_status_code(422):
        x.update_metadata(specs=has_repeated_spec)

    # A given spec cannot be too long.
    max_allowed_chars = ["a" * MAX_SPEC_CHARS]
    client.write_array([1, 2, 3], specs=max_allowed_chars)
    too_many_chars = ["a" * (1 + MAX_SPEC_CHARS)]
    with fail_with_status_code(422):
        client.write_array([1, 2, 3], specs=too_many_chars)
    with fail_with_status_code(422):
        x.update_metadata(specs=too_many_chars)

    # Up to 20 references are allowed.
    max_allowed_references = [
        {"label": f"ref{i}", "url": f"https://exmaple.com/{i}"}
        for i in range(MAX_ALLOWED_REFERENCES)
    ]
    y = client.write_array([1, 2, 3], references=max_allowed_references)
    y.update_metadata(references=max_allowed_references)  # no-op
    too_many_references = max_allowed_references + [
        {"label": "one_too_many", "url": "https://example.com/one_too_many"}
    ]
    with fail_with_status_code(422):
        client.write_array([1, 2, 3], references=too_many_references)
    with fail_with_status_code(422):
        y.update_metadata(references=too_many_references)

    # A given referenc label cannot be too long.
    max_allowed_chars = "a" * MAX_LABEL_CHARS
    client.write_array(
        [1, 2, 3],
        references=[{"label": max_allowed_chars, "url": "https://example.com"}],
    )
    too_many_chars = max_allowed_chars + "a"
    with fail_with_status_code(422):
        client.write_array(
            [1, 2, 3],
            references=[{"label": too_many_chars, "url": "https://example.com"}],
        )
    with fail_with_status_code(422):
        y.update_metadata(references=too_many_chars)



from dataclasses import dataclass
from typing import List, Tuple, Union


@dataclass
class DataFrameMicroStructure:
    meta: bytes  # Arrow-encoded DataFrame
    divisions: bytes  # Arrow-encoded DataFrame

    @classmethod
    def from_dask_dataframe(cls, ddf):
        # Make an *empty* DataFrame with the same structure as ddf.
        # TODO Look at make_meta_nonempty to see if the "objects" are str or
        # datetime or actually generic objects.
        import dask.dataframe.utils
        import pandas

        from ..serialization.dataframe import serialize_arrow

        meta = bytes(serialize_arrow(dask.dataframe.utils.make_meta(ddf), {}))
        divisions = bytes(
            serialize_arrow(pandas.DataFrame({"divisions": list(ddf.divisions)}), {})
        )
        return cls(meta=meta, divisions=divisions)

    @classmethod
    def from_dataframe(cls, df):
        # Make an *empty* DataFrame with the same structure as ddf.
        # TODO Look at make_meta_nonempty to see if the "objects" are str or
        # datetime or actually generic objects.
        import dask.dataframe
        import dask.dataframe.utils
        import pandas

        from ..serialization.dataframe import serialize_arrow

        ddf = dask.dataframe.from_pandas(df, npartitions=1)
        meta = bytes(serialize_arrow(dask.dataframe.utils.make_meta(ddf), {}))
        divisions = bytes(
            serialize_arrow(pandas.DataFrame({"divisions": list(ddf.divisions)}), {})
        )
        return cls(meta=meta, divisions=divisions)

    @property
    def meta_decoded(self):
        from ..serialization.dataframe import deserialize_arrow

        return deserialize_arrow(self.meta)

    @property
    def divisions_decoded(self):
        from ..serialization.dataframe import deserialize_arrow

        divisions_wrapped_in_df = deserialize_arrow(self.divisions)
        return tuple(divisions_wrapped_in_df["divisions"].values)


@dataclass
class DataFrameMacroStructure:
    npartitions: int
    columns: List[str]
    resizable: Union[bool, Tuple[bool, ...]] = False

    @classmethod
    def from_dask_dataframe(cls, ddf):
        return cls(npartitions=ddf.npartitions, columns=list(ddf.columns))


@dataclass
class DataFrameStructure:
    micro: DataFrameMicroStructure
    macro: DataFrameMacroStructure

    @classmethod
    def from_json(cls, content):
        return cls(
            micro=DataFrameMicroStructure(**content["micro"]),
            macro=DataFrameMacroStructure(**content["macro"]),
        )



import dask.array
import numpy
from dask.array.core import normalize_chunks

from ..server.object_cache import get_object_cache
from ..structures.array import ArrayMacroStructure, BuiltinDtype, StructDtype
from ..utils import DictView


class ArrayAdapter:
    """
    Wrap an array-like object in an interface that Tiled can serve.

    Examples
    --------

    Wrap any array-like.

    >>> ArrayAdapter(numpy.random.random((100, 100)))

    >>> ArrayAdapter(dask.array.from_array(numpy.random.random((100, 100)), chunks=(100, 50)))

    """

    structure_family = "array"

    def __init__(
        self,
        array,
        *,
        chunks=None,
        metadata=None,
        dims=None,
        specs=None,
        references=None,
    ):
        if chunks is None:
            if hasattr(array, "chunks"):
                chunks = array.chunks  # might be None
            else:
                chunks = None
            if chunks is None:
                chunks = ("auto",) * len(array.shape)
        self._chunks = normalize_chunks(
            chunks,
            shape=array.shape,
            dtype=array.dtype,
        )
        self._array = array
        self._metadata = metadata or {}
        self._dims = dims
        self.specs = specs or []
        self.references = references or []

    @classmethod
    def from_array(
        cls,
        array,
        *,
        chunks=None,
        metadata=None,
        dims=None,
        specs=None,
        references=None,
    ):
        return cls(
            numpy.asarray(array),
            chunks=chunks,
            metadata=metadata,
            dims=dims,
            specs=specs,
            references=references,
        )

    def __repr__(self):
        return f"{type(self).__name__}({self._array!r})"

    @property
    def metadata(self):
        return DictView(self._metadata)

    def macrostructure(self):
        "Structures of the layout of blocks of this array"
        return ArrayMacroStructure(
            shape=self._array.shape, chunks=self._chunks, dims=self._dims
        )

    def microstructure(self):
        "Internal structure of a block of this array --- i.e. its data type"
        if self._array.dtype.fields is not None:
            micro = StructDtype.from_numpy_dtype(self._array.dtype)
        else:
            micro = BuiltinDtype.from_numpy_dtype(self._array.dtype)
        return micro

    def read(self, slice=None):
        array = self._array
        if slice is not None:
            array = array[slice]
        # Special case for dask to cache computed result in object cache.
        if isinstance(self._array, dask.array.Array):
            # Note: If the cache is set to NO_CACHE, this is a null context.
            with get_object_cache().dask_context:
                return array.compute()
        return array

    def read_block(self, block, slice=None):
        # Slice the whole array to get this block.
        slice_, _ = slice_and_shape_from_block_and_chunks(block, self._chunks)
        array = self._array[slice_]
        # Slice within the block.
        if slice is not None:
            array = array[slice]
        # Special case for dask to cache computed result in object cache.
        if isinstance(array, dask.array.Array):
            # Note: If the cache is set to NO_CACHE, this is a null context.
            with get_object_cache().dask_context:
                return array.compute()
        return array


def slice_and_shape_from_block_and_chunks(block, chunks):
    """
    Given dask-like chunks and block id, return slice and shape of the block.
    """
    slice_ = []
    shape = []
    for b, c in zip(block, chunks):
        start = sum(c[:b])
        dim = c[b]
        slice_.append(slice(start, start + dim))
        shape.append(dim)
    return tuple(slice_), tuple(shape)



import io

from ..media_type_registration import serialization_registry
from ..utils import modules_available
from .dataframe import (
    APACHE_ARROW_FILE_MIME_TYPE,
    XLSX_MIME_TYPE,
    serialize_arrow,
    serialize_csv,
    serialize_excel,
    serialize_html,
    serialize_parquet,
)
from .node import walk


def as_dataset(node):
    import xarray

    data_vars = {}
    coords = {}
    for key, array_adapter in node.items():
        spec_names = set(spec.name for spec in array_adapter.specs)
        if "xarray_data_var" in spec_names:
            data_vars[key] = (
                array_adapter.macrostructure().dims,
                array_adapter.read(),
            )
        elif "xarray_coord" in spec_names:
            coords[key] = (
                array_adapter.macrostructure().dims,
                array_adapter.read(),
            )
        else:
            raise ValueError(
                "Child nodes of xarray_dataset should include spec "
                "'xarray_coord' or 'xarray_data_var'."
            )
    return xarray.Dataset(
        data_vars=data_vars, coords=coords, attrs=node.metadata["attrs"]
    )


class _BytesIOThatIgnoresClose(io.BytesIO):
    def close(self):
        # When the netcdf writer tells us to close(), ignore it.
        pass


if modules_available("scipy"):

    def serialize_netcdf(node, metadata):
        file = _BytesIOThatIgnoresClose()
        # Per the xarray.Dataset.to_netcdf documentation,
        # file-like objects are only supported by the scipy engine.
        as_dataset(node).to_netcdf(file, engine="scipy")
        return file.getbuffer()

    # Both application/netcdf and application/x-netcdf are used.
    # https://en.wikipedia.org/wiki/NetCDF
    serialization_registry.register(
        "xarray_dataset", "application/netcdf", serialize_netcdf
    )
    serialization_registry.register(
        "xarray_dataset", "application/x-netcdf", serialize_netcdf
    )

# Support DataFrame formats by first converting to DataFrame.
# This doesn't make much sense for N-dimensional variables, but for
# 1-dimensional variables it is useful.
serialization_registry.register(
    "xarray_dataset",
    APACHE_ARROW_FILE_MIME_TYPE,
    lambda node, metadata: serialize_arrow(as_dataset(node).to_dataframe(), metadata),
)
serialization_registry.register(
    "xarray_dataset",
    "application/x-parquet",
    lambda node, metadata: serialize_parquet(as_dataset(node).to_dataframe(), metadata),
)
serialization_registry.register(
    "xarray_dataset",
    "text/csv",
    lambda node, metadata: serialize_csv(as_dataset(node).to_dataframe(), metadata),
)
serialization_registry.register(
    "xarray_dataset",
    "text/x-comma-separated-values",
    lambda node, metadata: serialize_csv(as_dataset(node).to_dataframe(), metadata),
)
serialization_registry.register(
    "xarray_dataset",
    "text/plain",
    lambda node, metadata: serialize_csv(as_dataset(node).to_dataframe(), metadata),
)
serialization_registry.register(
    "xarray_dataset",
    "text/html",
    lambda node, metadata: serialize_html(as_dataset(node).to_dataframe(), metadata),
)
serialization_registry.register(
    "xarray_dataset",
    XLSX_MIME_TYPE,
    lambda node, metadata: serialize_excel(as_dataset(node).to_dataframe(), metadata),
)
if modules_available("orjson"):
    import orjson

    def serialize_json(node, metadata):
        df = as_dataset(node).to_dataframe()
        return orjson.dumps(
            {column: df[column].tolist() for column in df},
        )

    serialization_registry.register(
        "xarray_dataset",
        "application/json",
        serialize_json,
    )
if modules_available("h5py"):

    def serialize_hdf5(node, metadata):
        """
        Like for node, but encode everything under 'attrs' in attrs.
        """
        import h5py

        buffer = io.BytesIO()
        root_node = node
        with h5py.File(buffer, mode="w") as file:
            for k, v in metadata["attrs"].items():
                file.attrs.create(k, v)
            for key_path, array_adapter in walk(node):
                group = file
                node = root_node
                for key in key_path[:-1]:
                    node = node[key]
                    if key in group:
                        group = group[key]
                    else:
                        group = group.create_group(key)
                        group.attrs.update(node.metadata["attrs"])
                data = array_adapter.read()
                dataset = group.create_dataset(key_path[-1], data=data)
                for k, v in array_adapter.metadata["attrs"].items():
                    dataset.attrs.create(k, v)
        return buffer.getbuffer()

    serialization_registry.register(
        "xarray_dataset", "application/x-hdf5", serialize_hdf5
    )



"""
Test the feature wherein the descriptions (metadata, structure_family,
structure, etc.) for a node's children may be included with the
description of a node. This is used in xarray_dataset now and may
be used more widely later.
"""

import numpy
import xarray

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..adapters.xarray import DatasetAdapter
from ..client import from_tree, record_history
from ..server.core import INLINED_CONTENTS_LIMIT

tree = MapAdapter(
    {
        "dataset": DatasetAdapter.from_dataset(
            xarray.Dataset(
                data_vars={"temperature": ("time", [100, 99, 98])},
                coords={"time": [1, 2, 3]},
            )
        ),
    },
    metadata={"thing": "stuff"},
)
client = from_tree(tree)


def test_lookup():
    "Accessing a child node should trigger one request."
    with record_history() as history:
        client["dataset"]
    assert len(history.requests) == 1


def test_iter():
    "Iteration should be free because the contents were in-lined."
    expected = ["temperature", "time"]
    dsc = client["dataset"]
    with record_history() as history:
        assert list(dsc) == expected
    assert not history.requests

    # Implementation detail:
    # Without inlined contents, a request is needed.
    with record_history() as history:
        assert list(dsc.__iter__(_ignore_inlined_contents=True)) == expected
    assert history.requests


def test_keys_slice():
    "Iteration should be free because the contents were in-lined."
    expected = ["temperature", "time"]
    dsc = client["dataset"]
    with record_history() as history:
        assert list(dsc.keys()) == expected
    assert not history.requests

    with record_history() as history:
        assert dsc.keys()[:] == expected
    assert not history.requests

    with record_history() as history:
        assert dsc.keys().first() == "temperature"
        assert dsc.keys().last() == "time"
    assert not history.requests

    # Implementation detail:
    # Without inlined contents, a request is needed.
    with record_history() as history:
        list(dsc._keys_slice(0, 1, 1, _ignore_inlined_contents=True))
    assert history.requests


def test_items_slice():
    "Iteration should be free because the contents were in-lined."
    dsc = client["dataset"]
    with record_history() as history:
        list(dsc.items())
    assert not history.requests

    with record_history() as history:
        dsc.items()
    assert not history.requests

    with record_history() as history:
        dsc.items().first()
    assert not history.requests

    with record_history() as history:
        dsc.items().last()
    assert not history.requests

    # Implementation detail:
    # Without inlined contents, a request is needed.
    with record_history() as history:
        list(dsc._items_slice(0, 1, 1, _ignore_inlined_contents=True))
    assert history.requests


def test_too_wide_for_inline():
    """
    The server will not inline contents above a certain limit.


    It is fetched in pages on demand, as usual with nodes.
    """

    class Adapter(MapAdapter):
        def inlined_contents_enabled(self):
            return True

    a = numpy.array([1])
    tree = MapAdapter(
        {
            f"item{i:05}": ArrayAdapter(i * a)
            for i in range(1 + 2 * INLINED_CONTENTS_LIMIT)
        }
    )
    client = from_tree(tree)
    assert client.item["attributes"]["structure"]["contents"] is None
    with record_history() as history:
        assert set(client) == set(tree)
    assert len(history.requests) >= 3



from .queries import KeysFilter
from .scopes import SCOPES
from .utils import Sentinel, SpecialUsers, import_object

ALL_SCOPES = set(SCOPES)
NO_ACCESS = Sentinel("NO_ACCESS")


class DummyAccessPolicy:
    "Impose no access restrictions."

    def allowed_scopes(self, node, principal):
        return ALL_SCOPES

    def filters(self, node, principal, scopes):
        return []


class SimpleAccessPolicy:
    """
    A mapping of user names to lists of entries they have full access.

    This simple policy does not provide fine-grained control of scopes.

    >>> SimpleAccessPolicy({"alice": ["A", "B"], "bob": ["B"]}, provider="toy")
    """

    ALL = object()  # sentinel

    def __init__(self, access_lists, *, provider, scopes=None, public=None):
        self.access_lists = {}
        self.provider = provider
        self.scopes = scopes if (scopes is not None) else ALL_SCOPES
        self.public = set(public or [])
        for key, value in access_lists.items():
            if isinstance(value, str):
                value = import_object(value)
            self.access_lists[key] = value

    def _get_id(self, principal):
        # Get the id (i.e. username) of this Principal for the
        # associated authentication provider.
        for identity in principal.identities:
            if identity.provider == self.provider:
                id = identity.id
                break
        else:
            raise ValueError(
                f"Principcal {principal} has no identity from provider {self.provider}. "
                f"Its identities are: {principal.identities}"
            )
        return id

    def allowed_scopes(self, node, principal):
        # The simple policy does not provide for different Principals to
        # have different scopes on different Nodes. If the Principal has access,
        # they have the same hard-coded access everywhere.
        return self.scopes

    def filters(self, node, principal, scopes):
        if not scopes.issubset(self.scopes):
            return NO_ACCESS
        id = self._get_id(principal)
        access_list = self.access_lists.get(id, [])
        queries = []
        if not ((principal is SpecialUsers.admin) or (access_list is self.ALL)):
            allowed = set(access_list or []) | self.public
            queries.append(KeysFilter(allowed))
        return queries



from ..utils import modules_available


def register_builtin_serializers():
    """
    Register built-in serializers for each structure if its associated module(s) are installed.
    """
    # Each submodule in ..serialization registers serializers on import.
    # Some are conditional on the availability of particular I/O libraries.
    from ..serialization import node as _node  # noqa: F401

    del _node
    if modules_available("numpy", "dask.array"):
        from ..serialization import array as _array  # noqa: F401

        del _array
    if modules_available("pandas", "pyarrow", "dask.dataframe"):
        from ..serialization import dataframe as _dataframe  # noqa: F401

        del _dataframe
    if modules_available("sparse"):
        from ..serialization import sparse as _sparse  # noqa: F401

        del _sparse
    if modules_available("xarray"):
        from ..serialization import xarray as _xarray  # noqa: F401

        del _xarray



import contextlib
import getpass
import os
import re
import sys
import time
import urllib.parse
import warnings
from pathlib import Path

import httpx
import msgpack

from .._version import get_versions
from ..utils import DictView
from .auth import (
    DEFAULT_TOKEN_CACHE,
    CannotRefreshAuthentication,
    TiledAuth,
    build_refresh_request,
)
from .cache import Revalidate
from .utils import (
    DEFAULT_ACCEPTED_ENCODINGS,
    DEFAULT_TIMEOUT_PARAMS,
    EVENT_HOOKS,
    NotAvailableOffline,
    handle_error,
)

tiled_version = get_versions()["version"]
USER_AGENT = f"python-tiled/{tiled_version}"
API_KEY_AUTH_HEADER_PATTERN = re.compile(r"^Apikey (\w+)$")


class Context:
    """
    Wrap an httpx.Client with an optional cache and authentication functionality.
    """

    def __init__(
        self,
        uri,
        *,
        headers=None,
        api_key=None,
        cache=None,
        offline=False,
        timeout=None,
        verify=True,
        token_cache=DEFAULT_TOKEN_CACHE,
        app=None,
    ):
        # The uri is expected to reach the root API route.
        uri = httpx.URL(uri)
        headers = headers or {}
        headers.setdefault("accept-encoding", ",".join(DEFAULT_ACCEPTED_ENCODINGS))
        # Set the User Agent to help the server fail informatively if the client
        # version is too old.
        headers.setdefault("user-agent", USER_AGENT)

        # If ?api_key=... is present, move it from the query into a header.
        # The server would accept it in the query parameter, but using
        # a header is a little more secure (e.g. not logged).
        parsed_params = urllib.parse.parse_qs(uri.query.decode())
        api_key_list = parsed_params.pop("api_key", None)
        if api_key_list is not None:
            if api_key is not None:
                raise ValueError(
                    "api_key was provided as query parameter in URI and as keyword argument. Pick one."
                )
            if len(api_key_list) != 1:
                raise ValueError("Cannot handle two api_key query parameters")
            (api_key,) = api_key_list
        if api_key is None:
            # Check for an API key from the environment.
            api_key = os.getenv("TILED_API_KEY")
        # We will set the API key via the `api_key` property below,
        # after constructing the Client object.

        # FastAPI redirects /api -> /api/ so add it here to save a request.
        path = uri.path
        if not path.endswith("/"):
            path = f"{path}/"
        # Construct the uri *without* api_key param.
        # Drop any params/fragments.
        self.api_uri = httpx.URL(
            urllib.parse.urlunsplit((uri.scheme, uri.netloc.decode(), path, {}, ""))
        )
        if timeout is None:
            timeout = httpx.Timeout(**DEFAULT_TIMEOUT_PARAMS)
        if app is None:
            client = httpx.Client(
                verify=verify,
                event_hooks=EVENT_HOOKS,
                timeout=timeout,
                headers=headers,
                follow_redirects=True,
            )
        else:
            from ._testclient import TestClient

            # verify parameter is dropped, as there is no SSL in ASGI mode
            client = TestClient(
                app=app,
                event_hooks=EVENT_HOOKS,
                timeout=timeout,
                headers=headers,
            )
            client.follow_redirects = True
            client.__enter__()
            # The TestClient is meant to be used only as a context manager,
            # where the context starts and stops and the wrapped ASGI app.
            # We are abusing it slightly to enable interactive use of the
            # TestClient.
            #
            # Given a blank slate, we may not have chosen to support this at
            # all; we may insist on using this only within a context manager.
            # But for now it is necessary to suppot a smooth transition for
            # databroker. We may revisit it later.
            if sys.version_info < (3, 9):
                import atexit

                atexit.register(client.__exit__)
            else:
                # The threading module has its own (internal) atexit
                # mechanism that runs at thread shutdown, prior to the atexit
                # mechanism that runs at interpreter shutdown.
                # We need to intervene at that layer to close the portal, or else
                # we will wait forever for a thread run by the portal to join().
                import threading

                threading._register_atexit(client.__exit__)

        self.http_client = client
        self._verify = verify
        self._cache = cache
        self._revalidate = Revalidate.IF_WE_MUST
        self._offline = offline
        self._token_cache = Path(token_cache)

        # Make an initial "safe" request to:
        # (1) Get the server_info.
        # (2) Let the server set the CSRF cookie.
        # No authentication has been set up yet, so these requests will be unauthenticated.
        # https://cheatsheetseries.owasp.org/cheatsheets/Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html#double-submit-cookie
        if offline:
            self.server_info = self.get_json(self.api_uri)
        else:
            # We need a CSRF token.
            with self.disable_cache(allow_read=False, allow_write=True):
                self.server_info = self.get_json(self.api_uri)
        self.api_key = api_key  # property setter sets Authorization header

    def __getstate__(self):
        if getattr(self.http_client, "app", None):
            raise TypeError(
                "Cannot pickle a Tiled Context built around an ASGI. "
                "Only Tiled Context connected to remote servers can be pickled."
            )
        return (
            tiled_version,
            self.api_uri,
            self.http_client.headers,
            list(self.http_client.cookies.jar),
            self.http_client.timeout,
            self.http_client.auth,
            self._verify,
            self.offline,
            self._revalidate,
            self._token_cache,
            self.server_info,
            self.cache,
        )

    def __setstate__(self, state):
        (
            state_tiled_version,
            api_uri,
            headers,
            cookies_list,
            timeout,
            auth,
            verify,
            offline,
            revalidate,
            token_cache,
            server_info,
            cache,
        ) = state
        if state_tiled_version != tiled_version:
            raise TypeError(
                f"Cannot unpickle {type(self).__name__} from tiled version {state_tiled_version} "
                f"using tiled version {tiled_version}. Pickle should only be used to short-term "
                "transfer between identical versions of tiled."
            )
        self.api_uri = api_uri
        cookies = httpx.Cookies()
        for cookie in cookies_list:
            cookies.set(
                cookie.name, cookie.value, domain=cookie.domain, path=cookie.path
            )
        self.http_client = httpx.Client(
            verify=verify,
            event_hooks=EVENT_HOOKS,
            cookies=cookies,
            timeout=timeout,
            headers=headers,
            follow_redirects=True,
            auth=auth,
        )
        self._revalidate = revalidate
        self._token_cache = token_cache
        self._cache = cache
        self.server_info = server_info
        self.offline = offline

    @classmethod
    def from_any_uri(
        cls,
        uri,
        *,
        headers=None,
        api_key=None,
        cache=None,
        offline=False,
        timeout=None,
        verify=True,
        token_cache=DEFAULT_TOKEN_CACHE,
        app=None,
    ):
        """
        Accept a URI to a specific node.

        For example, given URI "https://example.com/api/v1/node/metadata/a/b/c"
        return a Context connected to "https://examples/api/v1" and the list
        ["a", "b", "c"].
        """
        uri = httpx.URL(uri)
        node_path_parts = []
        if "/node/metadata" in uri.path:
            api_path, _, node_path = uri.path.partition("/node/metadata")
            api_uri = uri.copy_with(path=api_path)
            node_path_parts.extend(
                [segment for segment in node_path.split("/") if segment]
            )
        # Below we handle the case where we are given *less* than the full URI
        # to the root endpoint. Here we are taking some care to plan for the case
        # where tiled is served at a sub-path, even though that is not yet supported
        # on the server side.
        elif "/api" not in uri.path:
            # Looks like we were given the root path (to the HTML landing page).
            path = uri.path
            if path.endswith("/"):
                path = path[:-1]
            api_uri = uri.copy_with(path=f"{path}/api/v1")
        elif "/v1" not in uri.path:
            # Looks like we were given the /api but no version.
            path = uri.path
            if path.endswith("/"):
                path = path[:-1]
            api_uri = uri.copy_with(path=f"{path}/v1")
        else:
            api_uri = uri
        context = cls(
            api_uri,
            headers=headers,
            api_key=api_key,
            cache=cache,
            offline=offline,
            timeout=timeout,
            verify=verify,
            token_cache=token_cache,
            app=app,
        )
        return context, node_path_parts

    @property
    def tokens(self):
        "A view of the current access and refresh tokens."
        return DictView(self.http_client.auth.tokens)

    @property
    def api_key(self):
        # Extract from header to ensure that there is one "ground truth" here
        # and no possibility of state getting out of sync.
        header = self.http_client.headers.get("Authorization", "")
        match = API_KEY_AUTH_HEADER_PATTERN.match(header)
        if match is not None:
            return match.group(1)

    @api_key.setter
    def api_key(self, api_key):
        if api_key is None:
            if self.http_client.headers.get("Authorization", "").startswith("Apikey "):
                self.http_client.headers.pop("Authorization")
        else:
            self.http_client.headers["Authorization"] = f"Apikey {api_key}"

    @property
    def cache(self):
        return self._cache

    @property
    def offline(self):
        return self._offline

    @offline.setter
    def offline(self, value):
        offline = bool(value)
        if offline and (self._cache is None):
            raise RuntimeError(
                """To use offline mode,  Tiled must be configured with a Cache, as in

>>> from tiled.client import from_uri
>>> from tiled.client.cache import Cache
>>> client = from_uri("...", cache=Cache.on_disk("my_cache"))
"""
            )
        self._offline = offline
        if not self._offline:
            # We need a CSRF token.
            with self.disable_cache(allow_read=False, allow_write=True):
                self.server_info = self.get_json(self.api_uri)

    def which_api_key(self):
        """
        A 'who am I' for API keys
        """
        if not self.api_key:
            raise RuntimeError("Not API key is configured for the client.")
        return self.get_json(self.server_info["authentication"]["links"]["apikey"])

    def create_api_key(self, scopes=None, expires_in=None, note=None):
        """
        Generate a new API for the currently-authenticated user.

        Parameters
        ----------
        scopes : Optional[List[str]]
            Restrict the access available to the API key by listing specific scopes.
            By default, this will have the same access as the user.
        expires_in : Optional[int]
            Number of seconds until API key expires. If None,
            it will never expire or it will have the maximum lifetime
            allowed by the server.
        note : Optional[str]
            Description (for humans).
        """
        return self.post_json(
            self.server_info["authentication"]["links"]["apikey"],
            {"scopes": scopes, "expires_in": expires_in, "note": note},
        )

    def revoke_api_key(self, first_eight):
        request = self.http_client.build_request(
            "DELETE",
            self.server_info["authentication"]["links"]["apikey"],
            headers={"x-csrf": self.http_client.cookies["tiled_csrf"]},
            params={"first_eight": first_eight},
        )
        response = self.http_client.send(request)
        handle_error(response)

    @property
    def app(self):
        warnings.warn(
            "The 'app' accessor on Context is deprecated. Use Context.http_client.app.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.http_client.app

    @property
    def base_url(self):
        warnings.warn(
            "The 'base_url' accessor on Context is deprecated. Use Context.http_client.base_url.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.http_client.base_url

    @property
    def event_hooks(self):
        "httpx.Client event hooks. This is exposed for testing."
        warnings.warn(
            "The 'event_hooks' accessor on Context is deprecated. Use Context.http_client.event_hooks.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.http_client.event_hooks

    @property
    def revalidate(self):
        """
        This controls how aggressively to check whether cache entries are out of date.

        - FORCE: Always revalidate (generally too aggressive and expensive)
        - IF_EXPIRED: Revalidate if the "Expire" date provided by the server has passed
        - IF_WE_MUST: Only revalidate if the server indicated that is is a
          particularly volatile entry, such as a search result to a dynamic query.
        """
        return self._revalidate

    @revalidate.setter
    def revalidate(self, value):
        self._revalidate = Revalidate(value)

    @contextlib.contextmanager
    def revalidation(self, revalidate):
        """
        Temporarily change the 'revalidate' property in a context.

        Parameters
        ----------
        revalidate: string or tiled.client.cache.Revalidate enum member
        """
        try:
            member = Revalidate(revalidate)
        except ValueError as err:
            # This raises a more helpful error that lists the valid options.
            raise ValueError(
                f"Revalidation {revalidate} not recognized. Must be one of {set(Revalidate.__members__)}"
            ) from err
        original = self.revalidate
        self.revalidate = member
        yield
        # Upon leaving context, set it back.
        self.revalidate = original

    @contextlib.contextmanager
    def disable_cache(self, allow_read=False, allow_write=False):
        self._disable_cache_read = not allow_read
        self._disable_cache_write = not allow_write
        yield
        self._disable_cache_read = False
        self._disable_cache_write = False

    def get_content(self, path, accept=None, stream=False, revalidate=None, **kwargs):
        send_kwargs = {}
        if "auth" in kwargs:
            send_kwargs["auth"] = kwargs.pop("auth")
        if revalidate is None:
            # Fallback to context default.
            revalidate = self.revalidate
        request = self.http_client.build_request("GET", path, **kwargs)
        if accept:
            request.headers["Accept"] = accept
        url = request.url
        if self._offline:
            # We must rely on the cache alone.
            # The role of a 'reservation' is to ensure that the content
            # of interest is not evicted from the cache between the moment
            # that we start verifying its validity and the moment that
            # we actually read the content. It is used more extensively
            # below.
            reservation = self._cache.get_reservation(url)
            if reservation is None:
                raise NotAvailableOffline(url)
            content = reservation.load_content()
            if content is None:
                # TODO Do we ever get here?
                raise NotAvailableOffline(url)
            return content
        if self._cache is None:
            # No cache, so we can use the client straightforwardly.
            response = self.http_client.send(request, stream=stream)
            handle_error(response)
            if response.headers.get("content-encoding") == "blosc":
                import blosc

                return blosc.decompress(response.content)
            return response.content
        # If we get this far, we have an online client and a cache.
        # Parse Cache-Control header directives.
        cache_control = {
            directive.lstrip(" ")
            for directive in request.headers.get("Cache-Control", "").split(",")
        }
        if "no-cache" in cache_control:
            reservation = None
        else:
            reservation = self._cache.get_reservation(url)
        try:
            if reservation is not None:
                is_stale = reservation.is_stale()
                if not (
                    # This condition means "client user wants us to unconditionally revalidate"
                    (revalidate == Revalidate.FORCE)
                    or
                    # This condition means "client user wants us to revalidate if expired"
                    (is_stale and (revalidate == Revalidate.IF_EXPIRED))
                    or
                    # This condition means "server really wants us to revalidate"
                    (is_stale and reservation.item.must_revalidate)
                    or self._disable_cache_read
                ):
                    # Short-circuit. Do not even bother consulting the server.
                    return reservation.load_content()
                if not self._disable_cache_read:
                    request.headers["If-None-Match"] = reservation.item.etag
            response = self.http_client.send(request, stream=stream, **send_kwargs)
            handle_error(response)
            if response.status_code == 304:  # HTTP 304 Not Modified
                # Update the expiration time.
                reservation.renew(response.headers.get("expires"))
                # Read from the cache
                return reservation.load_content()
            elif not response.is_error:
                etag = response.headers.get("ETag")
                encoding = response.headers.get("Content-Encoding")
                content = response.content
                # httpx handles standard HTTP encodings transparently, but we have to
                # handle "blosc" manually.
                if encoding == "blosc":
                    import blosc

                    content = blosc.decompress(content)
                if (
                    ("no-store" not in cache_control)
                    and (etag is not None)
                    and (not self._disable_cache_write)
                ):
                    # Write to cache.
                    self._cache.put(
                        url,
                        response.headers,
                        content,
                    )
                return content
            else:
                raise NotImplementedError(
                    f"Unexpected status_code {response.status_code}"
                )
        finally:
            if reservation is not None:
                reservation.ensure_released()

    def get_json(self, path, stream=False, **kwargs):
        return msgpack.unpackb(
            self.get_content(
                path, accept="application/x-msgpack", stream=stream, **kwargs
            ),
            timestamp=3,  # Decode msgpack Timestamp as datetime.datetime object.
        )

    def post_json(self, path, content):
        request = self.http_client.build_request(
            "POST",
            path,
            json=content,
            # Submit CSRF token in both header and cookie.
            # https://cheatsheetseries.owasp.org/cheatsheets/Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html#double-submit-cookie
            headers={
                "x-csrf": self.http_client.cookies["tiled_csrf"],
                "accept": "application/x-msgpack",
            },
        )
        response = self.http_client.send(request)
        handle_error(response)
        return msgpack.unpackb(
            response.content,
            timestamp=3,  # Decode msgpack Timestamp as datetime.datetime object.
        )

    def put_json(self, path, content):
        request = self.http_client.build_request(
            "PUT",
            path,
            json=content,
            # Submit CSRF token in both header and cookie.
            # https://cheatsheetseries.owasp.org/cheatsheets/Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html#double-submit-cookie
            headers={
                "x-csrf": self.http_client.cookies["tiled_csrf"],
                "accept": "application/x-msgpack",
            },
        )
        response = self.http_client.send(request)
        handle_error(response)
        return msgpack.unpackb(
            response.content,
            timestamp=3,  # Decode msgpack Timestamp as datetime.datetime object.
        )

    def put_content(self, path, content, headers=None, params=None):
        # Submit CSRF token in both header and cookie.
        # https://cheatsheetseries.owasp.org/cheatsheets/Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html#double-submit-cookie
        headers = headers or {}
        headers.setdefault("x-csrf", self.http_client.cookies["tiled_csrf"])
        headers.setdefault("accept", "application/x-msgpack")
        request = self.http_client.build_request(
            "PUT",
            path,
            content=content,
            headers=headers,
            params=params,
        )
        response = self.http_client.send(request)
        handle_error(response)
        return msgpack.unpackb(
            response.content,
            timestamp=3,  # Decode msgpack Timestamp as datetime.datetime object.
        )

    def delete_content(self, path, content, headers=None, params=None):
        # Submit CSRF token in both header and cookie.
        # https://cheatsheetseries.owasp.org/cheatsheets/Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html#double-submit-cookie
        headers = headers or {}
        headers.setdefault("x-csrf", self.http_client.cookies["tiled_csrf"])
        headers.setdefault("accept", "application/x-msgpack")
        request = self.http_client.build_request(
            "DELETE", path, content=None, headers=headers, params=params
        )
        response = self.http_client.send(request)
        handle_error(response)
        return msgpack.unpackb(
            response.content,
            timestamp=3,  # Decode msgpack Timestamp as datetime.datetime object.
        )

    def authenticate(self, username=None, provider=None):
        """
        See login. This is for programmatic use.
        """
        providers = self.server_info["authentication"]["providers"]
        spec = _choose_identity_provider(providers, provider)
        provider = spec["provider"]
        if self.api_key is not None:
            # Check that API key authenticates us as this user,
            # and then either return or raise.
            identities = self.whoami()["identities"]
            for identity in identities:
                if (identity["provider"] == provider) and (identity["id"] == username):
                    return
            raise RuntimeError(
                "An API key is set, and it is not associated with the username/provider "
                f"{username}/{provider}. Unset the API key first."
            )

        refresh_url = self.server_info["authentication"]["links"]["refresh_session"]
        csrf_token = self.http_client.cookies["tiled_csrf"]

        # If we are passed a username, we can check whether we already have
        # tokens stashed.
        if username is not None:
            token_directory = self._token_directory(provider, username)
            self.http_client.auth = TiledAuth(refresh_url, csrf_token, token_directory)
            # This will either:
            # * Use an access_token and succeed.
            # * Use a refresh_token to attempt refresh flow and succeed.
            # * Use a refresh_token to attempt refresh flow and fail, raise.
            # * Find no tokens and raise.
            try:
                self.whoami()
            except CannotRefreshAuthentication:
                # Continue below, where we will prompt for log in.
                self.http_client.auth = None
            else:
                # We have a live session for the specified provider and username already.
                # No need to log in again.
                return

        self.http_client.auth = None
        mode = spec["mode"]
        auth_endpoint = spec["links"]["auth_endpoint"]
        if mode == "password":
            if username:
                print(f"Username {username}")
            else:
                username = input("Username: ")
            password = getpass.getpass()
            form_data = {
                "grant_type": "password",
                "username": username,
                "password": password,
            }
            token_response = self.http_client.post(
                auth_endpoint, data=form_data, auth=None
            )
            handle_error(token_response)
            tokens = token_response.json()
        elif mode == "external":
            verification_response = self.http_client.post(
                auth_endpoint, json={}, auth=None
            )
            handle_error(verification_response)
            verification = verification_response.json()
            authorization_uri = verification["authorization_uri"]
            print(
                f"""
You have {int(verification['expires_in']) // 60} minutes visit this URL

{authorization_uri}

and enter the code: {verification['user_code']}

"""
            )
            import webbrowser

            webbrowser.open(authorization_uri)
            deadline = verification["expires_in"] + time.monotonic()
            print("Waiting...", end="", flush=True)
            while True:
                time.sleep(verification["interval"])
                if time.monotonic() > deadline:
                    raise Exception("Deadline expired.")
                access_response = self.http_client.post(
                    verification["verification_uri"],
                    json={
                        "device_code": verification["device_code"],
                        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                    },
                    auth=None,
                )
                if (access_response.status_code == 400) and (
                    access_response.json()["detail"]["error"] == "authorization_pending"
                ):
                    print(".", end="", flush=True)
                    continue
                handle_error(access_response)
                print("")
                break
            tokens = access_response.json()

        else:
            raise ValueError(f"Server has unknown authentication mechanism {mode!r}")
        username = tokens["identity"]["id"]
        token_directory = self._token_directory(provider, username)
        auth = TiledAuth(refresh_url, csrf_token, token_directory)
        auth.sync_set_token("access_token", tokens["access_token"])
        auth.sync_set_token("refresh_token", tokens["refresh_token"])
        self.http_client.auth = auth
        confirmation_message = spec.get("confirmation_message")
        if confirmation_message:
            print(confirmation_message.format(id=username))
        return spec, username

    def login(self, username=None, provider=None):
        """
        Depending on the server's authentication method, this will prompt for username/password:

        >>> c.login()
        Username: USERNAME
        Password: <input is hidden>

        or prompt you to open a link in a web browser to login with a third party:

        >>> c.login()
        You have ... minutes visit this URL

        https://...

        and enter the code: XXXX-XXXX
        """
        self.authenticate(username, provider)
        # For programmatic access to the return values, use authenticate().
        # This returns None in order to provide a clean UX in an interpreter.
        return None

    def _token_directory(self, provider, username):
        # ~/.config/tiled/tokens/{host:port}/{provider}/{username}
        # with each templated element URL-encoded so it is a valid filename.
        return Path(
            self._token_cache,
            urllib.parse.quote_plus(str(self.api_uri)),
            urllib.parse.quote_plus(provider),
            urllib.parse.quote_plus(username),
        )

    def force_auth_refresh(self):
        """
        Execute refresh flow.

        This method is exposed for testing and debugging uses.

        It should never be necessary for the user to call. Refresh flow is
        automatically executed by tiled.client.auth.TiledAuth when the current
        access_token expires.
        """
        if self.http_client.auth is None:
            raise RuntimeError(
                "No authentication has been set up. Cannot reauthenticate."
            )
        refresh_token = self.http_client.auth.sync_get_token(
            "refresh_token", reload_from_disk=True
        )
        if refresh_token is None:
            raise CannotRefreshAuthentication("There is no refresh_token.")
        csrf_token = self.http_client.cookies["tiled_csrf"]
        refresh_request = build_refresh_request(
            self.http_client.auth.refresh_url,
            refresh_token,
            csrf_token,
        )
        token_response = self.http_client.send(refresh_request, auth=None)
        if token_response.status_code == 401:
            raise CannotRefreshAuthentication(
                "Session cannot be refreshed. Log in again."
            )
        handle_error(token_response)
        tokens = token_response.json()
        self.http_client.auth.sync_set_token("access_token", tokens["access_token"])
        self.http_client.auth.sync_set_token("refresh_token", tokens["refresh_token"])
        return tokens

    def whoami(self):
        "Return information about the currently-authenticated user or service."
        return self.get_json(self.server_info["authentication"]["links"]["whoami"])

    def logout(self):
        """
        Log out of the current session (if any).

        This method is idempotent.
        """
        if self.http_client.auth is None:
            return

        # Revoke the current session.
        self.http_client.post(f"{self.api_uri}auth/logout")

        # Clear on-disk state.
        self.http_client.auth.sync_clear_token("access_token")
        self.http_client.auth.sync_clear_token("refresh_token")

        # Clear in-memory state.
        self.http_client.headers.pop("Authorization", None)
        self.http_client.auth = None

    def revoke_session(self, session_id):
        """
        Revoke a Session so it cannot be refreshed.

        This may be done to ensure that a possibly-leaked refresh token cannot be used.
        """
        request = self.http_client.build_request(
            "DELETE",
            self.server_info["authentication"]["links"]["revoke_session"].format(
                session_id=session_id
            ),
            headers={"x-csrf": self.http_client.cookies["tiled_csrf"]},
        )
        response = self.http_client.send(request)
        handle_error(response)


def _choose_identity_provider(providers, provider=None):
    if provider is not None:
        for spec in providers:
            if spec["provider"] == provider:
                break
        else:
            raise ValueError(
                f"No such provider {provider}. Choices are {[spec['provider'] for spec in providers]}"
            )
    else:
        if len(providers) == 1:
            # There is only one choice, so no need to prompt the user.
            (spec,) = providers
        else:
            while True:
                print("Authenticaiton providers:")
                for i, spec in enumerate(providers, start=1):
                    print(f"{i} - {spec['provider']}")
                raw_choice = input(
                    "Choose an authentication provider (or press Enter to escape): "
                )
                if not raw_choice:
                    print("No authentication provider chosen. Failed.")
                    break
                try:
                    choice = int(raw_choice)
                except TypeError:
                    print("Choice must be a number.")
                    continue
                try:
                    spec = providers[choice - 1]
                except IndexError:
                    print(f"Choice must be a number 1 through {len(providers)}.")
                    continue
                break
    return spec



import os
from pathlib import Path

import appdirs
import typer

CLI_CACHE_DIR = Path(
    os.getenv(
        "TILED_CLI_CACHE_DIR", os.path.join(appdirs.user_cache_dir("tiled"), "cli")
    )
)


def get_profile(name):
    from ..profiles import load_profiles

    profiles = load_profiles()
    if name is None:
        # Use the default profile.
        # Raise if it is not set or if it is set but does not exit.
        filepath = CLI_CACHE_DIR / "active_profile"
        if not filepath.is_file():
            typer.echo(
                """No default profile set. Use:

    tiled connect ...

to set a default profile or else specify a profile for this command using

    tiled ... --profile=PROFILE
""",
                err=True,
            )
            raise typer.Abort()
        with open(filepath, "r") as file:
            name = file.read()
        if name not in profiles:
            filepath.unlink()
            typer.echo(
                f"""Default profile {name!r} does not exist. Clearing default. Use:

    tiled connect ...

to set a default profile or else specify a profile for this command using

    tiled ... --profile=PROFILE
""",
                err=True,
            )
            raise typer.Abort()
    try:
        _, profile = profiles[name]
        if "direct" in profile:
            typer.echo(
                f"Profile {profile!r} uses in a direct (in-process) Tiled server "
                "and cannot be connected to from the CLI.",
                err=True,
            )
            typer.Abort()
    except KeyError:
        typer.echo(
            f"""Profile {name!r} could not be found. Use:

    tiled profile list

to list choices.""",
            err=True,
        )
        raise typer.Abort()
    return name, profile


def get_default_profile_name():
    filepath = CLI_CACHE_DIR / "active_profile"
    if not filepath.is_file():
        return None
    with open(filepath, "r") as file:
        return file.read()


def get_context(profile):
    import json

    from ..client.context import Context

    profile_name, profile_content = get_profile(profile)
    filepath = CLI_CACHE_DIR / "profile_auths" / profile_name
    context, _ = Context.from_any_uri(
        profile_content["uri"], verify=profile_content.get("verify", True)
    )
    if filepath.is_file():
        with open(filepath, "r") as file:
            auth = json.load(file)
        context.authenticate(auth["username"], auth["provider"])
    return context



import uuid

import numpy as np
import pandas as pd

from ..adapters.array import ArrayAdapter
from ..adapters.dataframe import DataFrameAdapter
from ..adapters.mapping import MapAdapter
from ..client import from_tree

values = ["a", "b", "c"]
counts = [10, 5, 7]

mapping = {}

for value, count in zip(values, counts):
    for _ in range(count):
        mapping[str(uuid.uuid4())] = MapAdapter({}, metadata={"foo": {"bar": value}})

# items which do not contain the queries metadata should not effect the results
for _ in range(10):
    mapping[str(uuid.uuid4())] = ArrayAdapter.from_array(
        np.ones(10), metadata={}, specs=["MyArray"]
    )

for _ in range(10):
    mapping[str(uuid.uuid4())] = DataFrameAdapter.from_pandas(
        pd.DataFrame({"a": np.ones(10)}),
        metadata={},
        specs=["MyDataFrame"],
        npartitions=1,
    )

tree = MapAdapter(mapping)
client = from_tree(tree)


def test_distinct():
    # test without counts
    distinct = client.distinct(
        "foo.bar", structure_families=True, specs=True, counts=False
    )
    expected = {
        "metadata": {"foo.bar": [{"value": v, "count": None} for v in values]},
        "specs": [
            {"value": [], "count": None},
            {"value": ["MyArray"], "count": None},
            {"value": ["MyDataFrame"], "count": None},
        ],
        "structure_families": [
            {"value": "node", "count": None},
            {"value": "array", "count": None},
            {"value": "dataframe", "count": None},
        ],
    }

    assert distinct["metadata"] == expected["metadata"]
    assert distinct["specs"] == expected["specs"]
    assert distinct["structure_families"] == expected["structure_families"]

    # test with counts
    distinct = client.distinct(
        "foo.bar", structure_families=True, specs=True, counts=True
    )
    expected = {
        "metadata": {
            "foo.bar": [{"value": v, "count": c} for v, c in zip(values, counts)]
        },
        "specs": [
            {"value": [], "count": 22},
            {"value": ["MyArray"], "count": 10},
            {"value": ["MyDataFrame"], "count": 10},
        ],
        "structure_families": [
            {"value": "node", "count": 22},
            {"value": "array", "count": 10},
            {"value": "dataframe", "count": 10},
        ],
    }

    assert distinct["metadata"] == expected["metadata"]
    assert distinct["specs"] == expected["specs"]
    assert distinct["structure_families"] == expected["structure_families"]

    # test with no matches
    distinct = client.distinct("baz", counts=True)
    expected = {"baz": []}
    assert distinct["metadata"] == expected



import base64
import dataclasses
import inspect
from datetime import datetime, timedelta
from typing import Any, List, Optional

from fastapi import APIRouter, Depends, HTTPException, Query, Request, Security
from jmespath.exceptions import JMESPathError
from pydantic import BaseSettings

from .. import __version__
from ..structures.core import StructureFamily
from ..validation_registration import ValidationError
from . import schemas
from .authentication import Mode, get_authenticators, get_current_principal
from .core import (
    DEPTH_LIMIT,
    NoEntry,
    UnsupportedMediaTypes,
    WrongTypeForRoute,
    construct_data_response,
    construct_entries_response,
    construct_resource,
    construct_revisions_response,
    json_or_msgpack,
    resolve_media_type,
)
from .dependencies import (
    SecureEntry,
    block,
    expected_shape,
    get_query_registry,
    get_serialization_registry,
    get_validation_registry,
    slice_,
)
from .settings import get_settings
from .utils import (
    FilteredNode,
    filter_for_access,
    get_base_url,
    get_structure,
    record_timing,
)

DEFAULT_PAGE_SIZE = 100
MAX_PAGE_SIZE = 300


router = APIRouter()


@router.get("/", response_model=schemas.About)
async def about(
    request: Request,
    settings: BaseSettings = Depends(get_settings),
    authenticators=Depends(get_authenticators),
    serialization_registry=Depends(get_serialization_registry),
    query_registry=Depends(get_query_registry),
    # This dependency is here because it runs the code that moves
    # API key from the query parameter to a cookie (if it is valid).
    principal=Security(get_current_principal, scopes=[]),
):
    # TODO The lazy import of entry modules and serializers means that the
    # lists of formats are not populated until they are first used. Not very
    # helpful for discovery! The registration can be made non-lazy, while the
    # imports of the underlying I/O libraries themselves (openpyxl, pillow,
    # etc.) can remain lazy.
    request.state.endpoint = "about"
    base_url = get_base_url(request)
    authentication = {
        "required": not settings.allow_anonymous_access,
    }
    provider_specs = []
    for provider, authenticator in authenticators.items():
        if authenticator.mode == Mode.password:
            spec = {
                "provider": provider,
                "mode": authenticator.mode.value,
                "links": {
                    "auth_endpoint": f"{base_url}/auth/provider/{provider}/token"
                },
                "confirmation_message": getattr(
                    authenticator, "confirmation_message", None
                ),
            }
        elif authenticator.mode == Mode.external:
            spec = {
                "provider": provider,
                "mode": authenticator.mode.value,
                "links": {
                    "auth_endpoint": f"{base_url}/auth/provider/{provider}/authorize"
                },
                "confirmation_message": getattr(
                    authenticator, "confirmation_message", None
                ),
            }
        else:
            # It should be impossible to reach here.
            assert False
        provider_specs.append(spec)
    if provider_specs:
        # If there are *any* authenticaiton providers, these
        # endpoints will be added.
        authentication["links"] = {
            "whoami": f"{base_url}/auth/whoami",
            "apikey": f"{base_url}/auth/apikey",
            "refresh_session": f"{base_url}/auth/session/refresh",
            "revoke_session": f"{base_url}/auth/session/revoke/{{session_id}}",
            "logout": f"{base_url}/auth/logout",
        }
    authentication["providers"] = provider_specs

    return json_or_msgpack(
        request,
        schemas.About(
            library_version=__version__,
            api_version=0,
            formats={
                structure_family: list(
                    serialization_registry.media_types(structure_family)
                )
                for structure_family in serialization_registry.structure_families
            },
            aliases={
                structure_family: serialization_registry.aliases(structure_family)
                for structure_family in serialization_registry.structure_families
            },
            queries=list(query_registry.name_to_query_type),
            authentication=authentication,
            links={
                "self": base_url,
                "documentation": f"{base_url}docs",
            },
            meta={"root_path": request.scope.get("root_path") or "" + "/api"},
        ).dict(),
        expires=datetime.utcnow() + timedelta(seconds=600),
    )


def declare_search_router(query_registry):
    """
    This is done dynamically at router startup.

    We check the registry of known search query types, which is user
    configurable, and use that to define the allowed HTTP query parameters for
    this route.
    """

    async def node_search(
        request: Request,
        path: str,
        fields: Optional[List[schemas.EntryFields]] = Query(list(schemas.EntryFields)),
        select_metadata: Optional[str] = Query(None),
        offset: Optional[int] = Query(0, alias="page[offset]", ge=0),
        limit: Optional[int] = Query(
            DEFAULT_PAGE_SIZE, alias="page[limit]", ge=0, le=MAX_PAGE_SIZE
        ),
        sort: Optional[str] = Query(None),
        max_depth: Optional[int] = Query(None, ge=0, le=DEPTH_LIMIT),
        omit_links: bool = Query(False),
        entry: Any = SecureEntry(scopes=["read:metadata"]),
        query_registry=Depends(get_query_registry),
        principal: str = Depends(get_current_principal),
        **filters,
    ):
        request.state.endpoint = "search"
        if entry.structure_family != "node":
            raise WrongTypeForRoute(
                "This is not a Node; it cannot be searched or listed."
            )
        entry = filter_for_access(
            entry, principal, ["read:metadata"], request.state.metrics
        )
        try:
            resource, metadata_stale_at, must_revalidate = construct_entries_response(
                query_registry,
                entry,
                "/node/search",
                path,
                offset,
                limit,
                fields,
                select_metadata,
                omit_links,
                filters,
                sort,
                get_base_url(request),
                resolve_media_type(request),
                max_depth=max_depth,
            )
            # We only get one Expires header, so if different parts
            # of this response become stale at different times, we
            # cite the earliest one.
            entries_stale_at = getattr(entry, "entries_stale_at", None)
            headers = {}
            if (metadata_stale_at is None) or (entries_stale_at is None):
                expires = None
            else:
                expires = min(metadata_stale_at, entries_stale_at)
            if must_revalidate:
                headers["Cache-Control"] = "must-revalidate"
            return json_or_msgpack(
                request,
                resource.dict(),
                expires=expires,
                headers=headers,
            )
        except NoEntry:
            raise HTTPException(status_code=404, detail="No such entry.")
        except WrongTypeForRoute as err:
            raise HTTPException(status_code=404, detail=err.args[0])
        except JMESPathError as err:
            raise HTTPException(
                status_code=400,
                detail=f"Malformed 'select_metadata' parameter raised JMESPathError: {err}",
            )

    # Black magic here! FastAPI bases its validation and auto-generated swagger
    # documentation on the signature of the route function. We do not know what
    # that signature should be at compile-time. We only know it once we have a
    # chance to check the user-configurable registry of query types. Therefore,
    # we modify the signature here, at runtime, just before handing it to
    # FastAPI in the usual way.

    # When FastAPI calls the function with these added parameters, they will be
    # accepted via **filters.

    # Make a copy of the original parameters.
    signature = inspect.signature(node_search)
    parameters = list(signature.parameters.values())
    # Drop the **filters parameter from the signature.
    del parameters[-1]
    # Add a parameter for each field in each type of query.
    for name, query in query_registry.name_to_query_type.items():
        for field in dataclasses.fields(query):
            # The structured "alias" here is based on
            # https://mglaman.dev/blog/using-json-router-query-your-search-router-indexes
            if getattr(field.type, "__origin__", None) is list:
                field_type = str
            else:
                field_type = field.type
            injected_parameter = inspect.Parameter(
                name=f"filter___{name}___{field.name}",
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=Query(None, alias=f"filter[{name}][condition][{field.name}]"),
                annotation=Optional[List[field_type]],
            )
            parameters.append(injected_parameter)
    node_search.__signature__ = signature.replace(parameters=parameters)
    # End black magic

    # Register the search route.
    router = APIRouter()
    router.get(
        "/node/search",
        response_model=schemas.Response[
            List[schemas.Resource[schemas.NodeAttributes, dict, dict]],
            schemas.PaginationLinks,
            dict,
        ],
        include_in_schema=False,
    )(node_search)
    router.get(
        "/node/search/{path:path}",
        response_model=schemas.Response[
            List[schemas.Resource[schemas.NodeAttributes, dict, dict]],
            schemas.PaginationLinks,
            dict,
        ],
    )(node_search)
    return router


@router.get("/node/distinct/{path:path}", response_model=schemas.GetDistinctResponse)
async def node_distinct(
    request: Request,
    structure_families: bool = False,
    specs: bool = False,
    metadata: Optional[List[str]] = Query(default=[]),
    counts: bool = False,
    entry: Any = SecureEntry(scopes=["read:metadata"]),
):
    if hasattr(entry, "get_distinct"):
        distinct = entry.get_distinct(
            metadata=metadata,
            structure_families=structure_families,
            specs=specs,
            counts=counts,
        )
        return json_or_msgpack(
            request, schemas.GetDistinctResponse.parse_obj(distinct).dict()
        )
    else:
        raise HTTPException(
            status_code=405, detail="This path does not support distinct."
        )


@router.get(
    "/node/metadata/{path:path}",
    response_model=schemas.Response[
        schemas.Resource[schemas.NodeAttributes, dict, dict], dict, dict
    ],
)
async def node_metadata(
    request: Request,
    path: str,
    fields: Optional[List[schemas.EntryFields]] = Query(list(schemas.EntryFields)),
    select_metadata: Optional[str] = Query(None),
    max_depth: Optional[int] = Query(None, ge=0, le=DEPTH_LIMIT),
    omit_links: bool = Query(False),
    entry: Any = SecureEntry(scopes=["read:metadata"]),
    root_path: bool = Query(False),
):
    "Fetch the metadata and structure information for one entry."

    request.state.endpoint = "metadata"
    base_url = get_base_url(request)
    path_parts = [segment for segment in path.split("/") if segment]
    try:
        resource = construct_resource(
            base_url,
            path_parts,
            entry,
            fields,
            select_metadata,
            omit_links,
            resolve_media_type(request),
            max_depth=max_depth,
        )
    except JMESPathError as err:
        raise HTTPException(
            status_code=400,
            detail=f"Malformed 'select_metadata' parameter raised JMESPathError: {err}",
        )
    meta = {"root_path": request.scope.get("root_path") or "/"} if root_path else {}
    return json_or_msgpack(
        request,
        schemas.Response(data=resource, meta=meta).dict(),
        expires=getattr(entry, "metadata_stale_at", None),
    )


@router.get(
    "/array/block/{path:path}", response_model=schemas.Response, name="array block"
)
def array_block(
    request: Request,
    entry=SecureEntry(scopes=["read:data"]),
    block=Depends(block),
    slice=Depends(slice_),
    expected_shape=Depends(expected_shape),
    format: Optional[str] = None,
    filename: Optional[str] = None,
    serialization_registry=Depends(get_serialization_registry),
    settings: BaseSettings = Depends(get_settings),
):
    """
    Fetch a chunk of array-like data.
    """
    if entry.structure_family not in {"array", "sparse"}:
        raise HTTPException(
            status_code=404,
            detail=f"Cannot read {entry.structure_family} structure with /array/block route.",
        )
    if block == ():
        # Handle special case of numpy scalar.
        if entry.macrostructure().shape != ():
            raise HTTPException(
                status_code=400,
                detail=f"Requested scalar but shape is {entry.macrostructure().shape}",
            )
        with record_timing(request.state.metrics, "read"):
            array = entry.read()
    else:
        try:
            with record_timing(request.state.metrics, "read"):
                array = entry.read_block(block, slice=slice)
        except IndexError:
            raise HTTPException(status_code=400, detail="Block index out of range")
        if (expected_shape is not None) and (expected_shape != array.shape):
            raise HTTPException(
                status_code=400,
                detail=f"The expected_shape {expected_shape} does not match the actual shape {array.shape}",
            )
    if array.nbytes > settings.response_bytesize_limit:
        raise HTTPException(
            status_code=400,
            detail=(
                f"Response would exceed {settings.response_bytesize_limit}. "
                "Use slicing ('?slice=...') to request smaller chunks."
            ),
        )
    try:
        with record_timing(request.state.metrics, "pack"):
            return construct_data_response(
                entry.structure_family,
                serialization_registry,
                array,
                entry.metadata,
                request,
                format,
                specs=getattr(entry, "specs", []),
                expires=getattr(entry, "content_stale_at", None),
                filename=filename,
            )
    except UnsupportedMediaTypes as err:
        # raise HTTPException(status_code=406, detail=", ".join(err.supported))
        raise HTTPException(status_code=406, detail=err.args[0])


@router.get(
    "/array/full/{path:path}", response_model=schemas.Response, name="full array"
)
def array_full(
    request: Request,
    entry=SecureEntry(scopes=["read:data"]),
    slice=Depends(slice_),
    expected_shape=Depends(expected_shape),
    format: Optional[str] = None,
    filename: Optional[str] = None,
    serialization_registry=Depends(get_serialization_registry),
    settings: BaseSettings = Depends(get_settings),
):
    """
    Fetch a slice of array-like data.
    """
    structure_family = entry.structure_family
    if structure_family not in {"array", "sparse"}:
        raise HTTPException(
            status_code=404,
            detail=f"Cannot read {entry.structure_family} structure with /array/full route.",
        )
    # Deferred import because this is not a required dependency of the server
    # for some use cases.
    import numpy

    try:
        with record_timing(request.state.metrics, "read"):
            array = entry.read(slice)
        if structure_family == "array":
            array = numpy.asarray(array)  # Force dask or PIMS or ... to do I/O.
    except IndexError:
        raise HTTPException(status_code=400, detail="Block index out of range")
    if (expected_shape is not None) and (expected_shape != array.shape):
        raise HTTPException(
            status_code=400,
            detail=f"The expected_shape {expected_shape} does not match the actual shape {array.shape}",
        )
    if array.nbytes > settings.response_bytesize_limit:
        raise HTTPException(
            status_code=400,
            detail=(
                "Response would exceed {settings.response_bytesize_limit}. "
                "Use slicing ('?slice=...') to request smaller chunks."
            ),
        )
    try:
        with record_timing(request.state.metrics, "pack"):
            return construct_data_response(
                structure_family,
                serialization_registry,
                array,
                entry.metadata,
                request,
                format,
                specs=getattr(entry, "specs", []),
                expires=getattr(entry, "content_stale_at", None),
                filename=filename,
            )
    except UnsupportedMediaTypes as err:
        raise HTTPException(status_code=406, detail=err.args[0])


@router.get(
    "/dataframe/partition/{path:path}",
    response_model=schemas.Response,
    name="dataframe partition",
)
def dataframe_partition(
    request: Request,
    partition: int,
    entry=SecureEntry(scopes=["read:data"]),
    field: Optional[List[str]] = Query(None, min_length=1),
    format: Optional[str] = None,
    filename: Optional[str] = None,
    serialization_registry=Depends(get_serialization_registry),
    settings: BaseSettings = Depends(get_settings),
):
    """
    Fetch a partition (continuous block of rows) from a DataFrame.
    """
    if entry.structure_family != "dataframe":
        raise HTTPException(
            status_code=404,
            detail=f"Cannot read {entry.structure_family} structure with /dataframe/partition route.",
        )
    try:
        # The singular/plural mismatch here of "fields" and "field" is
        # due to the ?field=A&field=B&field=C... encodes in a URL.
        with record_timing(request.state.metrics, "read"):
            df = entry.read_partition(partition, fields=field)
    except IndexError:
        raise HTTPException(status_code=400, detail="Partition out of range")
    except KeyError as err:
        (key,) = err.args
        raise HTTPException(status_code=400, detail=f"No such field {key}.")
    if df.memory_usage().sum() > settings.response_bytesize_limit:
        raise HTTPException(
            status_code=400,
            detail=(
                "Response would exceed {settings.response_bytesize_limit}. "
                "Select a subset of the columns ('?field=...') to "
                "request a smaller chunks."
            ),
        )
    try:
        with record_timing(request.state.metrics, "pack"):
            return construct_data_response(
                "dataframe",
                serialization_registry,
                df,
                entry.metadata,
                request,
                format,
                specs=getattr(entry, "specs", []),
                expires=getattr(entry, "content_stale_at", None),
                filename=filename,
            )
    except UnsupportedMediaTypes as err:
        raise HTTPException(status_code=406, detail=err.args[0])


@router.get(
    "/node/full/{path:path}",
    response_model=schemas.Response,
    name="full generic 'node' or 'dataframe'",
)
def node_full(
    request: Request,
    entry=SecureEntry(scopes=["read:data"]),
    principal: str = Depends(get_current_principal),
    field: Optional[List[str]] = Query(None, min_length=1),
    format: Optional[str] = None,
    filename: Optional[str] = None,
    serialization_registry=Depends(get_serialization_registry),
    settings: BaseSettings = Depends(get_settings),
):
    """
    Fetch the data below the given node.
    """
    try:
        # The singular/plural mismatch here of "fields" and "field" is
        # due to the ?field=A&field=B&field=C... encodes in a URL.
        with record_timing(request.state.metrics, "read"):
            data = entry.read(fields=field)
    except KeyError as err:
        (key,) = err.args
        raise HTTPException(status_code=400, detail=f"No such field {key}.")
    if (entry.structure_family == "dataframe") and (
        data.memory_usage().sum() > settings.response_bytesize_limit
    ):
        raise HTTPException(
            status_code=400,
            detail=(
                "Response would exceed {settings.response_bytesize_limit}. "
                "Select a subset of the columns ('?field=...') to "
                "request a smaller chunks."
            ),
        )
    if entry.structure_family == "node":
        data = FilteredNode(data, principal, ["read:data"], request.state.metrics)
        # TODO Walk node to determine size before handing off to serializer.
    try:
        with record_timing(request.state.metrics, "pack"):
            return construct_data_response(
                entry.structure_family,
                serialization_registry,
                data,
                entry.metadata,
                request,
                format,
                specs=getattr(entry, "specs", []),
                expires=getattr(entry, "content_stale_at", None),
                filename=filename,
            )
    except UnsupportedMediaTypes as err:
        raise HTTPException(status_code=406, detail=err.args[0])


@router.post("/node/metadata/{path:path}", response_model=schemas.PostMetadataResponse)
def post_metadata(
    request: Request,
    path: str,
    body: schemas.PostMetadataRequest,
    validation_registry=Depends(get_validation_registry),
    entry=SecureEntry(scopes=["write:metadata", "create"]),
):
    if not hasattr(entry, "post_metadata"):
        raise HTTPException(status_code=405, detail="This path cannot accept metadata.")

    if body.structure_family == StructureFamily.dataframe:
        # Decode meta for pydantic validation
        body.structure.micro.meta = base64.b64decode(body.structure.micro.meta)
        body.structure.micro.divisions = base64.b64decode(
            body.structure.micro.divisions
        )

    metadata, structure_family, structure, specs, references = (
        body.metadata,
        body.structure_family,
        body.structure,
        body.specs,
        body.references,
    )

    metadata_modified = False

    # Specs should be ordered from most specific/constrained to least.
    # Validate them in reverse order, with the least constrained spec first,
    # because it may do normalization that helps pass the more constrained one.
    # Known Issue:
    # When there is more than one spec, it's possible for the validator for
    # Spec 2 to make a modification that breaks the validation for Spec 1.
    # For now we leave it to the server maintainer to ensure that validators
    # won't step on each other in this way, but this may need revisiting.
    for spec in reversed(specs):
        if spec.name not in validation_registry:
            raise HTTPException(
                status_code=400, detail=f"Unrecognized spec: {spec.name}"
            )
        validator = validation_registry(spec.name)
        try:
            result = validator(metadata, structure_family, structure, spec, references)
        except ValidationError as e:
            raise HTTPException(
                status_code=400, detail=f"failed validation for spec {spec.name}:\n{e}"
            )
        if result is not None:
            metadata_modified = True
            metadata = result

    key = entry.post_metadata(
        metadata=body.metadata,
        structure_family=body.structure_family,
        structure=body.structure,
        specs=body.specs,
        references=body.references,
    )
    links = {}
    base_url = get_base_url(request)
    path_parts = [segment for segment in path.split("/") if segment] + [key]
    path_str = "/".join(path_parts)
    links["self"] = f"{base_url}/node/metadata/{path_str}"
    if body.structure_family == StructureFamily.array:
        block_template = ",".join(
            f"{{{index}}}" for index in range(len(body.structure.macro.shape))
        )
        links["block"] = f"{base_url}/array/block/{path_str}?block={block_template}"
        links["full"] = f"{base_url}/array/full/{path_str}"
    elif body.structure_family == StructureFamily.sparse:
        # Different from array because of structure.macro.shape vs structure.shape
        # Can be unified if we drop macro/micro namespace.
        block_template = ",".join(
            f"{{{index}}}" for index in range(len(body.structure.shape))
        )
        links["block"] = f"{base_url}/array/block/{path_str}?block={block_template}"
        links["full"] = f"{base_url}/array/full/{path_str}"
    elif body.structure_family == StructureFamily.dataframe:
        links[
            "partition"
        ] = f"{base_url}/dataframe/partition/{path_str}?partition={{index}}"
        links["full"] = f"{base_url}/node/full/{path_str}"
    else:
        raise NotImplementedError(body.structure_family)
    response_data = {"id": key, "links": links}
    if metadata_modified:
        response_data["metadata"] = metadata
    return json_or_msgpack(request, response_data)


@router.delete("/node/metadata/{path:path}")
async def delete(
    request: Request,
    entry=SecureEntry(scopes=["write:data", "write:metadata"]),
):
    if hasattr(entry, "delete"):
        entry.delete()
    else:
        raise HTTPException(
            status_code=405, detail="This path does not support deletion."
        )
    return json_or_msgpack(request, None)


@router.put("/array/full/{path:path}")
async def put_array_full(
    request: Request,
    entry=SecureEntry(scopes=["write:data"]),
):
    data = await request.body()
    if hasattr(entry, "put_data"):
        entry.put_data(data)
    else:
        raise HTTPException(
            status_code=405, detail="This path cannot accept array data."
        )
    return json_or_msgpack(request, None)


@router.put("/array/block/{path:path}")
async def put_array_block(
    request: Request,
    entry=SecureEntry(scopes=["write:data"]),
    block=Depends(block),
):
    data = await request.body()

    if hasattr(entry, "put_data"):
        entry.put_data(data, block=block)
    else:
        raise HTTPException(
            status_code=405, detail="This path cannot accept array data."
        )
    return json_or_msgpack(request, None)


@router.put("/node/full/{path:path}")
async def put_dataframe_full(
    request: Request,
    entry=SecureEntry(scopes=["write:data"]),
):
    data = await request.body()

    if hasattr(entry, "put_data"):
        entry.put_data(data)
    else:
        raise HTTPException(
            status_code=405, detail="This path cannot accept dataframe data."
        )
    return json_or_msgpack(request, None)


@router.put("/dataframe/partition/{path:path}")
async def put_dataframe_partition(
    partition: int,
    request: Request,
    entry=SecureEntry(scopes=["write:data"]),
):
    data = await request.body()

    if hasattr(entry, "put_data"):
        entry.put_data(data, partition=partition)
    else:
        raise HTTPException(
            status_code=405, detail="This path cannot accept dataframe data."
        )
    return json_or_msgpack(request, None)


@router.put("/node/metadata/{path:path}")
async def put_metadata(
    request: Request,
    body: schemas.PutMetadataRequest,
    validation_registry=Depends(get_validation_registry),
    entry=SecureEntry(scopes=["write:metadata"]),
):
    if not hasattr(entry, "put_metadata"):
        raise HTTPException(
            status_code=405, detail="This path does not support update of metadata."
        )

    metadata, structure_family, structure, specs, references = (
        body.metadata if body.metadata is not None else entry.metadata,
        entry.structure_family,
        get_structure(entry),
        body.specs if body.specs is not None else entry.specs,
        body.references if body.references is not None else entry.references,
    )

    metadata_modified = False

    # Specs should be ordered from most specific/constrained to least.
    # Validate them in reverse order, with the least constrained spec first,
    # because it may do normalization that helps pass the more constrained one.
    # Known Issue:
    # When there is more than one spec, it's possible for the validator for
    # Spec 2 to make a modification that breaks the validation for Spec 1.
    # For now we leave it to the server maintainer to ensure that validators
    # won't step on each other in this way, but this may need revisiting.
    for spec in reversed(specs):
        if spec.name not in validation_registry:
            raise HTTPException(
                status_code=400, detail=f"Unrecognized spec: {spec.name}"
            )
        validator = validation_registry(spec.name)
        try:
            result = validator(metadata, structure_family, structure, spec, references)
        except ValidationError as e:
            raise HTTPException(
                status_code=400,
                detail=f"failed validation for spec {spec.name}:\n{e}",
            )
        if result is not None:
            metadata_modified = True
            metadata = result

    entry.put_metadata(metadata=metadata, specs=specs, references=references)

    response_data = {"id": entry.key}
    if metadata_modified:
        response_data["metadata"] = metadata
    return json_or_msgpack(request, response_data)


@router.get("/node/revisions/{path:path}")
async def node_revisions(
    request: Request,
    path: str,
    offset: Optional[int] = Query(0, alias="page[offset]", ge=0),
    limit: Optional[int] = Query(
        DEFAULT_PAGE_SIZE, alias="page[limit]", ge=0, le=MAX_PAGE_SIZE
    ),
    entry=SecureEntry(scopes=["read:metadata"]),
):
    if not hasattr(entry, "revisions"):
        raise HTTPException(
            status_code=405, detail="This path does not support update of metadata."
        )

    base_url = get_base_url(request)
    resource = construct_revisions_response(
        entry,
        base_url,
        "/node/revisions",
        path,
        offset,
        limit,
        resolve_media_type(request),
    )
    return json_or_msgpack(request, resource.dict())


@router.delete("/node/revisions/{path:path}")
async def revisions_delitem(
    request: Request, n: int, entry=SecureEntry(scopes=["write:metadata"])
):
    if not hasattr(entry, "revisions"):
        raise HTTPException(
            status_code=405,
            detail="This path does not support a del request for revisions.",
        )

    entry.revisions.delete_revision(n)
    return json_or_msgpack(request, None)



from pathlib import Path
from typing import Optional

import typer

admin_app = typer.Typer()


@admin_app.command("initialize-database")
def initialize_database(database_uri: str):
    """
    Initialize a SQL database for use by Tiled.
    """
    from sqlalchemy import create_engine

    from ..database.core import (
        REQUIRED_REVISION,
        UninitializedDatabase,
        check_database,
        initialize_database,
    )

    engine = create_engine(database_uri)
    redacted_url = engine.url._replace(password="[redacted]")
    try:
        check_database(engine)
    except UninitializedDatabase:
        # Create tables and stamp (alembic) revision.
        typer.echo(
            f"Database {redacted_url} is new. Creating tables and marking revision {REQUIRED_REVISION}.",
            err=True,
        )
        initialize_database(engine)
        typer.echo("Database initialized.", err=True)
    else:
        typer.echo(f"Database at {redacted_url} is already initialized.", err=True)
        raise typer.Abort()


@admin_app.command("upgrade-database")
def upgrade_database(
    database_uri: str,
    revision: Optional[str] = typer.Argument(
        None,
        help="The ID of a revision to upgrade to. By default, upgrade to the latest one.",
    ),
):
    """
    Upgrade the database schema to the latest version.
    """
    from sqlalchemy import create_engine

    from ..database.core import get_current_revision, upgrade

    engine = create_engine(database_uri)
    redacted_url = engine.url._replace(password="[redacted]")
    current_revision = get_current_revision(engine)
    if current_revision is None:
        # Create tables and stamp (alembic) revision.
        typer.echo(
            f"Database {redacted_url} has not been initialized. Use `tiled admin initialize-database`.",
            err=True,
        )
        raise typer.Abort()
    upgrade(engine, revision or "head")


@admin_app.command("downgrade-database")
def downgrade_database(
    database_uri: str,
    revision: str = typer.Argument(..., help="The ID of a revision to downgrade to."),
):
    """
    Upgrade the database schema to the latest version.
    """
    from sqlalchemy import create_engine

    from ..database.core import downgrade, get_current_revision

    engine = create_engine(database_uri)
    redacted_url = engine.url._replace(password="[redacted]")
    current_revision = get_current_revision(engine)
    if current_revision is None:
        # Create tables and stamp (alembic) revision.
        typer.echo(
            f"Database {redacted_url} has not been initialized. Use `tiled admin initialize-database`.",
            err=True,
        )
        raise typer.Abort()
    downgrade(engine, revision)


@admin_app.command("check-config")
def check_config(
    config_path: Path = typer.Argument(
        None,
        help=(
            "Path to a config file or directory of config files. "
            "If None, check environment variable TILED_CONFIG. "
            "If that is unset, try default location ./config.yml."
        ),
    ),
):
    "Check configuration file for syntax and validation errors."
    import os

    from ..config import parse_configs

    config_path = config_path or os.getenv("TILED_CONFIG", "config.yml")
    try:
        parse_configs(config_path)
    except Exception as err:
        typer.echo(str(err), err=True)
        raise typer.Exit(1)
    typer.echo("No errors found in configuration.")



import io
import json

import numpy
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..client import from_config
from .utils import fail_with_status_code
from .writable_adapters import WritableMapAdapter

arr = ArrayAdapter.from_array(numpy.ones((5, 5)))


def tree_a(access_policy=None):
    return MapAdapter({"A1": arr, "A2": arr}, access_policy=access_policy)


def tree_b(access_policy=None):
    return MapAdapter({"B1": arr, "B2": arr}, access_policy=access_policy)


def writable_tree(access_policy=None):
    return WritableMapAdapter({"A1": arr, "A2": arr}, access_policy=access_policy)


@pytest.fixture
def config(tmpdir):
    return {
        "authentication": {
            "secret_keys": ["SECRET"],
            "providers": [
                {
                    "provider": "toy",
                    "authenticator": "tiled.authenticators:DictionaryAuthenticator",
                    "args": {
                        "users_to_passwords": {"alice": "secret1", "bob": "secret2"}
                    },
                }
            ],
        },
        "database": {
            "uri": f"sqlite:///{tmpdir}/tiled.sqlite",
        },
        "access_control": {
            "access_policy": "tiled.access_policies:SimpleAccessPolicy",
            "args": {
                "access_lists": {"alice": ["a", "c", "d", "e"]},
                "provider": "toy",
            },
        },
        "trees": [
            {
                "tree": f"{__name__}:tree_a",
                "path": "/a",
                "access_control": {
                    "access_policy": "tiled.access_policies:SimpleAccessPolicy",
                    "args": {
                        "provider": "toy",
                        "access_lists": {
                            "alice": ["A2"],
                            # This should have no effect because bob
                            # cannot access the parent node.
                            "bob": ["A1", "A2"],
                        },
                    },
                },
            },
            {"tree": f"{__name__}:tree_b", "path": "/b", "access_policy": None},
            {
                "tree": f"{__name__}:writable_tree",
                "path": "/c",
                "access_control": {
                    "access_policy": "tiled.access_policies:SimpleAccessPolicy",
                    "args": {
                        "provider": "toy",
                        "access_lists": {
                            "alice": "tiled.access_policies:SimpleAccessPolicy.ALL",
                        },
                    },
                },
            },
            {
                "tree": f"{__name__}:writable_tree",
                "path": "/d",
                "access_control": {
                    "access_policy": "tiled.access_policies:SimpleAccessPolicy",
                    "args": {
                        "provider": "toy",
                        "access_lists": {
                            "alice": "tiled.access_policies:SimpleAccessPolicy.ALL",
                        },
                        # Block writing.
                        "scopes": ["read:metadata", "read:data"],
                    },
                },
            },
            {
                "tree": f"{__name__}:writable_tree",
                "path": "/e",
                "access_control": {
                    "access_policy": "tiled.access_policies:SimpleAccessPolicy",
                    "args": {
                        "provider": "toy",
                        "access_lists": {
                            "alice": "tiled.access_policies:SimpleAccessPolicy.ALL",
                        },
                        # Block creation.
                        "scopes": [
                            "read:metadata",
                            "read:data",
                            "write:metadata",
                            "write:data",
                        ],
                    },
                },
            },
        ],
    }


def test_top_level_access_control(enter_password, config, tmpdir):
    with enter_password("secret1"):
        alice_client = from_config(config, username="alice", token_cache=tmpdir)
    with enter_password("secret2"):
        bob_client = from_config(config, username="bob", token_cache=tmpdir)
    assert "a" in alice_client
    assert "A2" in alice_client["a"]
    assert "A1" not in alice_client["a"]
    assert "b" not in alice_client
    alice_client["a"]["A2"]
    with pytest.raises(KeyError):
        alice_client["b"]
    assert not list(bob_client)
    with pytest.raises(KeyError):
        bob_client["a"]
    with pytest.raises(KeyError):
        bob_client["b"]


def test_node_export(enter_password, config, tmpdir):
    "Exporting a node should include only the children we can see."
    with enter_password("secret1"):
        alice_client = from_config(config, username="alice", token_cache=tmpdir)
    buffer = io.BytesIO()
    alice_client.export(buffer, format="application/json")
    buffer.seek(0)
    exported_dict = json.loads(buffer.read())
    assert "a" in exported_dict["contents"]
    assert "A2" in exported_dict["contents"]["a"]["contents"]
    assert "A1" not in exported_dict["contents"]["a"]["contents"]
    assert "b" not in exported_dict
    exported_dict["contents"]["a"]["contents"]["A2"]


def test_create_and_update_allowed(enter_password, config, tmpdir):
    with enter_password("secret1"):
        alice_client = from_config(config, username="alice", token_cache=tmpdir)

    # Update
    alice_client["c"].metadata
    alice_client["c"].update_metadata(metadata={"added_key": 3})
    assert alice_client["c"].metadata["added_key"] == 3

    # Create
    alice_client["c"].write_array([1, 2, 3])


def test_writing_blocked_by_access_policy(enter_password, config, tmpdir):
    with enter_password("secret1"):
        alice_client = from_config(config, username="alice", token_cache=tmpdir)

    alice_client["d"].metadata
    with fail_with_status_code(403):
        alice_client["d"].update_metadata(metadata={"added_key": 3})


def test_create_blocked_by_access_policy(enter_password, config, tmpdir):
    with enter_password("secret1"):
        alice_client = from_config(config, username="alice", token_cache=tmpdir)

    with fail_with_status_code(403):
        alice_client["e"].write_array([1, 2, 3])



from typing import List, Tuple, Union

import pandas
from pydantic import BaseModel

from ..media_type_registration import deserialization_registry
from ..serialization.dataframe import serialize_arrow
from ..utils import APACHE_ARROW_FILE_MIME_TYPE


class DataFrameMicroStructure(BaseModel):
    meta: bytes
    divisions: bytes

    @classmethod
    def from_dask_dataframe(cls, ddf):
        # Make an *empty* DataFrame with the same structure as ddf.
        # TODO Look at make_meta_nonempty to see if the "objects" are str or
        # datetime or actually generic objects.
        import dask.dataframe.utils

        meta = bytes(serialize_arrow(dask.dataframe.utils.make_meta(ddf), {}))
        divisions = bytes(
            serialize_arrow(pandas.DataFrame({"divisions": list(ddf.divisions)}), {})
        )
        return cls(meta=meta, divisions=divisions)

    @classmethod
    def from_dataframe(cls, df):
        # Make an *empty* DataFrame with the same structure as ddf.
        # TODO Look at make_meta_nonempty to see if the "objects" are str or
        # datetime or actually generic objects.
        import dask.dataframe
        import dask.dataframe.utils

        ddf = dask.dataframe.from_pandas(df, npartitions=1)
        meta = bytes(serialize_arrow(dask.dataframe.utils.make_meta(ddf), {}))
        divisions = bytes(
            serialize_arrow(pandas.DataFrame({"divisions": list(ddf.divisions)}), {})
        )
        return cls(meta=meta, divisions=divisions)


class DataFrameMacroStructure(BaseModel):
    npartitions: int
    columns: List[str]
    resizable: Union[bool, Tuple[bool, ...]] = False

    @classmethod
    def from_dask_dataframe(cls, ddf):
        return cls(npartitions=ddf.npartitions, columns=list(ddf.columns))


class DataFrameStructure(BaseModel):
    micro: DataFrameMicroStructure
    macro: DataFrameMacroStructure

    @classmethod
    def from_json(cls, content):
        divisions_wrapped_in_df = deserialization_registry(
            "dataframe", APACHE_ARROW_FILE_MIME_TYPE, content["micro"]["divisions"]
        )
        divisions = tuple(divisions_wrapped_in_df["divisions"].values)
        meta = deserialization_registry(
            "dataframe", APACHE_ARROW_FILE_MIME_TYPE, content["micro"]["meta"]
        )
        return cls(
            micro=DataFrameMicroStructure(meta=meta, divisions=divisions),
            macro=DataFrameMacroStructure(**content["macro"]),
        )



import uuid

import dask
import numpy

from ..adapters.array import ArrayAdapter, slice_and_shape_from_block_and_chunks
from ..adapters.dataframe import DataFrameAdapter
from ..adapters.mapping import MapAdapter
from ..adapters.sparse import COOAdapter
from ..serialization.dataframe import deserialize_arrow
from ..structures.core import Spec, StructureFamily


class _WritableMixin:
    def __init__(self, *args, key="", specs=None, **kwargs):
        self.key = key
        specs = [Spec(**dict(s)) for s in (specs or [])]
        super().__init__(*args, specs=specs, **kwargs)

    def put_metadata(self, metadata, specs, references):
        # TODO This skips over validation and has a race condition in it, but
        # this test harness is not long for this world anyway, so good enough
        # for now.
        self._metadata.clear()
        self._metadata.update(metadata)
        self.specs.clear()
        self.specs.extend(Spec(**dict(s)) for s in specs)
        self.references.clear()
        self.references.extend(references)


class WritableArrayAdapter(_WritableMixin, ArrayAdapter):
    def put_data(self, body, block=None):
        macrostructure = self.macrostructure()
        if block is None:
            shape = macrostructure.shape
            slice_ = numpy.s_[:]
        else:
            slice_, shape = slice_and_shape_from_block_and_chunks(
                block, macrostructure.chunks
            )
        array = numpy.frombuffer(
            body, dtype=self.microstructure().to_numpy_dtype()
        ).reshape(shape)
        self._array[slice_] = array


class WritableDataFrameAdapter(_WritableMixin, DataFrameAdapter):
    def put_data(self, body, partition=0):
        df = deserialize_arrow(body)
        self._partitions[partition] = df


class WritableCOOAdapter(_WritableMixin, COOAdapter):
    def put_data(self, body, block=None):
        if not block:
            block = (0,) * len(self.shape)
        df = deserialize_arrow(body)
        coords = df[df.columns[:-1]].values.T
        data = df["data"].values
        self.blocks[block] = (coords, data)


class WritableMapAdapter(_WritableMixin, MapAdapter):
    def post_metadata(self, metadata, structure_family, structure, specs, references):
        key = str(uuid.uuid4())
        if structure_family == StructureFamily.array:
            # Initialize an array of zeros, similar to how chunked storage
            # formats (e.g. HDF5, Zarr) use a fill_value.
            array = dask.array.zeros(
                structure.macro.shape,
                dtype=structure.micro.to_numpy_dtype(),
                chunks=structure.macro.chunks,
            )
            self._mapping[key] = WritableArrayAdapter(
                array,
                metadata=metadata,
                specs=specs,
                key=key,
                references=references,
            )
        elif structure_family == StructureFamily.dataframe:
            # Initialize an empty DataFrame with the right columns/types.
            meta = deserialize_arrow(structure.micro.meta)
            divisions_wrapped_in_df = deserialize_arrow(structure.micro.divisions)
            divisions = tuple(divisions_wrapped_in_df["divisions"].values)
            self._mapping[key] = WritableDataFrameAdapter(
                [None] * structure.macro.npartitions,
                meta=meta,
                divisions=divisions,
                metadata=metadata,
                specs=specs,
                key=key,
                references=references,
            )
        elif structure_family == StructureFamily.sparse:
            self._mapping[key] = WritableCOOAdapter(
                {},
                shape=structure.shape,
                chunks=structure.chunks,
                metadata=metadata,
                specs=specs,
                key=key,
                references=references,
            )
        else:
            raise NotImplementedError(structure_family)
        return key



from pathlib import Path
from typing import Optional

import typer

serve_app = typer.Typer()


@serve_app.command("directory")
def serve_directory(
    directory: str,
    public: bool = typer.Option(
        False,
        "--public",
        help=(
            "Turns off requirement for API key authentication for reading. "
            "However, the API key is still required for writing, so data cannot be modified even with "
            "this option selected."
        ),
    ),
    keep_ext: bool = typer.Option(
        False,
        "--keep-ext",
        help=(
            "Serve a file like 'measurements.csv' as its full filepath with extension, "
            "instead of the default which would serve it as 'measurements'. "
            "This is discouraged because it leaks details about the storage "
            "format to the client, such that changing the storage in the future "
            "may break user (client-side) code."
        ),
    ),
    poll_interval: float = typer.Option(
        None,
        "--poll-interval",
        help=(
            "Time in seconds between scans of the directory for removed or "
            "changed files. If 0, do not poll for changes."
        ),
    ),
    host: str = typer.Option(
        "127.0.0.1",
        help=(
            "Bind socket to this host. Use `--host 0.0.0.0` to make the application "
            "available on your local network. IPv6 addresses are supported, for "
            "example: --host `'::'`."
        ),
    ),
    port: int = typer.Option(8000, help="Bind to a socket with this port."),
    object_cache_available_bytes: Optional[float] = typer.Option(
        None,
        "--data-cache",
        help=(
            "Maximum size for the object cache, given as a number of bytes as in "
            "1_000_000 or as a fraction of system RAM (total physical memory) as in "
            "0.3. Set to 0 to disable this cache. By default, it will use up to "
            "0.15 (15%) of RAM."
        ),
    ),
):
    "Serve a Tree instance from a directory of files."
    from ..adapters.files import DirectoryAdapter
    from ..server.app import build_app, print_admin_api_key_if_generated

    tree_kwargs = {}
    server_settings = {}
    if keep_ext:
        from ..adapters.files import identity

        tree_kwargs.update({"key_from_filename": identity})
    if poll_interval is not None:
        tree_kwargs.update({"poll_interval": poll_interval})
    if object_cache_available_bytes is not None:
        server_settings["object_cache"] = {}
        server_settings["object_cache"][
            "available_bytes"
        ] = object_cache_available_bytes
    tree = DirectoryAdapter.from_directory(directory, **tree_kwargs)
    web_app = build_app(tree, {"allow_anonymous_access": public}, server_settings)
    print_admin_api_key_if_generated(web_app, host=host, port=port)

    import uvicorn

    uvicorn.run(web_app, host=host, port=port)


@serve_app.command("pyobject")
def serve_pyobject(
    object_path: str = typer.Argument(
        ..., help="Object path, as in 'package.subpackage.module:object_name'"
    ),
    public: bool = typer.Option(
        False,
        "--public",
        help=(
            "Turns off requirement for API key authentication for reading. "
            "However, the API key is still required for writing, so data cannot be modified even with this "
            "option selected."
        ),
    ),
    host: str = typer.Option(
        "127.0.0.1",
        help=(
            "Bind socket to this host. Use `--host 0.0.0.0` to make the application "
            "available on your local network. IPv6 addresses are supported, for "
            "example: --host `'::'`."
        ),
    ),
    port: int = typer.Option(8000, help="Bind to a socket with this port."),
    object_cache_available_bytes: Optional[float] = typer.Option(
        None,
        "--data-cache",
        help=(
            "Maximum size for the object cache, given as a number of bytes as in "
            "1_000_000 or as a fraction of system RAM (total physical memory) as in "
            "0.3. Set to 0 to disable this cache. By default, it will use up to "
            "0.15 (15%) of RAM."
        ),
    ),
):
    "Serve a Tree instance from a Python module."
    from ..server.app import build_app, print_admin_api_key_if_generated
    from ..utils import import_object

    tree = import_object(object_path)
    server_settings = {}
    if object_cache_available_bytes is not None:
        server_settings["object_cache"] = {}
        server_settings["object_cache"][
            "available_bytes"
        ] = object_cache_available_bytes
    web_app = build_app(tree, {"allow_anonymous_access": public}, server_settings)
    print_admin_api_key_if_generated(web_app, host=host, port=port)

    import uvicorn

    uvicorn.run(web_app, host=host, port=port)


@serve_app.command("config")
def serve_config(
    config_path: Path = typer.Argument(
        None,
        help=(
            "Path to a config file or directory of config files. "
            "If None, check environment variable TILED_CONFIG. "
            "If that is unset, try default location ./config.yml."
        ),
    ),
    public: bool = typer.Option(
        False,
        "--public",
        help=(
            "Turns off requirement for API key authentication for reading. "
            "However, the API key is still required for writing, so data cannot be modified even with this "
            "option selected."
        ),
    ),
    host: str = typer.Option(
        None,
        help=(
            "Bind socket to this host. Use `--host 0.0.0.0` to make the application "
            "available on your local network. IPv6 addresses are supported, for "
            "example: --host `'::'`. Uses value in config by default."
        ),
    ),
    port: int = typer.Option(
        None, help="Bind to a socket with this port. Uses value in config by default."
    ),
):
    "Serve a Tree as specified in configuration file(s)."
    import os

    from ..config import construct_build_app_kwargs, parse_configs

    config_path = config_path or os.getenv("TILED_CONFIG", "config.yml")
    try:
        parsed_config = parse_configs(config_path)
    except Exception as err:
        typer.echo(str(err), err=True)
        raise typer.Abort()

    # Let --public flag override config.
    if public:
        if "authentication" not in parsed_config:
            parsed_config["authentication"] = {}
        parsed_config["authentication"]["allow_anonymous_access"] = True

    # Delay this import so that we can fail faster if config-parsing fails above.

    from ..server.app import build_app, logger, print_admin_api_key_if_generated

    # Extract config for uvicorn.
    uvicorn_kwargs = parsed_config.pop("uvicorn", {})
    # If --host is given, it overrides host in config. Same for --port.
    uvicorn_kwargs["host"] = host or uvicorn_kwargs.get("host", "127.0.0.1")
    uvicorn_kwargs["port"] = port or uvicorn_kwargs.get("port", 8000)

    # This config was already validated when it was parsed. Do not re-validate.
    kwargs = construct_build_app_kwargs(parsed_config, source_filepath=config_path)
    logger.info(f"Using configuration from {Path(config_path).absolute()}")
    web_app = build_app(**kwargs)
    print_admin_api_key_if_generated(
        web_app, host=uvicorn_kwargs["host"], port=uvicorn_kwargs["port"]
    )

    # Likewise, delay this import.

    import uvicorn

    uvicorn.run(web_app, **uvicorn_kwargs)



import builtins
import hashlib

import tifffile

from ..server.object_cache import with_object_cache
from ..structures.array import ArrayMacroStructure, BuiltinDtype


class TiffAdapter:
    """
    Read a TIFF file.

    Examples
    --------

    >>> TiffAdapter("path/to/file.tiff")
    """

    structure_family = "array"

    def __init__(self, path, *, specs=None, references=None):
        self._file = tifffile.TiffFile(path)
        self._cache_key = (type(self).__module__, type(self).__qualname__, path)
        self.specs = specs or []
        self.references = references or []

    @property
    def metadata(self):
        # This contains some enums, but Python's built-in JSON serializer
        # handles them fine (converting  to str or int as appropriate).
        return {tag.name: tag.value for tag in self._file.pages[0].tags.values()}

    def read(self, slice=None):
        # TODO Is there support for reading less than the whole array
        # if we only want a slice? I do not think that is possible with a
        # single-page TIFF but I'm not sure. Certainly it *is* possible for
        # multi-page TIFFs.
        arr = with_object_cache(self._cache_key, self._file.asarray)
        if slice is not None:
            arr = arr[slice]
        return arr

    def read_block(self, block, slice=None):
        # For simplicity, this adapter always treat a single TIFF file as one
        # chunk. This could be relaxed in the future.
        if sum(block) != 0:
            raise IndexError(block)

        arr = with_object_cache(self._cache_key, self._file.asarray)
        if slice is not None:
            arr = arr[slice]
        return arr

    def microstructure(self):
        return BuiltinDtype.from_numpy_dtype(self._file.series[0].dtype)

    def macrostructure(self):
        if self._file.is_shaped:
            shape = tuple(self._file.shaped_metadata[0]["shape"])
        else:
            arr = with_object_cache(self._cache_key, self._file.asarray)
            shape = arr.shape
        return ArrayMacroStructure(shape=shape, chunks=tuple((dim,) for dim in shape))


class TiffSequenceAdapter:

    structure_family = "array"

    def __init__(self, seq, *, specs=None, references=None):
        self._seq = seq
        self._metadata = {}
        self._cache_key = (
            type(self).__module__,
            type(self).__qualname__,
            hashlib.md5(str(seq.files).encode()).hexdigest(),
        )
        self.specs = specs or []
        self.references = references or []

    @property
    def metadata(self):
        return self._metadata

    def read(self, slice=None):
        """Return a numpy array

        Receives a sequence of values to select from a collection of tiff files that were saved in a folder
        The input order is defined as files --> X slice --> Y slice
        read() can receive one value or one slice to select all the data from one file or a sequence of files;
        or it can receive a tuple of up to three values (int or slice) to select a more specific sequence of pixels
        of a group of images
        """

        # Print("Inside Adapter:", slice)
        if slice is None:
            return with_object_cache(self._cache_key, self._seq.asarray)
        if isinstance(slice, int):
            # e.g. read(slice=0)
            return with_object_cache(
                self._cache_key + (slice,),
                tifffile.TiffFile(self._seq.files[slice]).asarray,
            )
        # e.g. read(slice=(...))
        if isinstance(slice, tuple):
            if len(slice) == 0:
                return with_object_cache(self._cache_key, self._seq.asarray)
            image_axis, *the_rest = slice
            # Could be int or slice
            # (0, slice(...)) or (0,....) are converted to a list
            if isinstance(image_axis, int):
                # e.g. read(slice=(0, ....))
                arr = with_object_cache(
                    self._cache_key + (image_axis,),
                    tifffile.TiffFile(self._seq.files[image_axis]).asarray,
                )
            if isinstance(image_axis, builtins.slice):
                if image_axis.start is None:
                    slice_start = 0
                else:
                    slice_start = image_axis.start
                if image_axis.step is None:
                    slice_step = 1
                else:
                    slice_step = image_axis.step

                arr = with_object_cache(
                    self._cache_key,
                    tifffile.TiffSequence(
                        self._seq.files[
                            slice_start : image_axis.stop : slice_step  # noqa: E203
                        ]
                    ).asarray,
                )
            arr = arr[tuple(the_rest)]
            return arr
        if isinstance(slice, builtins.slice):
            # Check for start and step which can be optional
            if slice.start is None:
                slice_start = 0
            else:
                slice_start = slice.start
            if slice.step is None:
                slice_step = 1
            else:
                slice_step = slice.step

            arr = with_object_cache(
                self._cache_key,
                tifffile.TiffSequence(
                    self._seq.files[slice_start : slice.stop : slice_step]  # noqa: E203
                ).asarray,
            )
            return arr

    def read_block(self, block, slice=None):
        if block[1:] != (0, 0):
            raise IndexError(block)
        arr = self.read(builtins.slice(block[0], block[0] + 1))
        if slice is not None:
            arr = arr[slice]
        return arr

    def microstructure(self):
        # Assume all files have the same data type
        return BuiltinDtype.from_numpy_dtype(self.read(slice=0).dtype)

    def macrostructure(self):
        shape = (len(self._seq), *self.read(slice=0).shape)
        return ArrayMacroStructure(
            shape=shape,
            # one chunks per underlying TIFF file
            chunks=((1,) * shape[0], (shape[1],), (shape[2],)),
        )


def subdirectory_handler(path):
    """
    Sniff a subdirectory for TIFF sequences.
    """
    # Max fraction of files that do not look like TIFF sequence files
    RELATIVE_THRESHOLD = 0.2
    # Max number of files that do not look like TIFF sequence files
    ABSOLUTE_THRESHOLD = 10
    filepaths = []
    outliers = 0
    for filepath in path.iterdir():
        if not filepath.is_file():
            # Skip directories.
            continue
        if filepath.name.startswith("."):
            # Skip hidden files.
            continue
        if (filepath.suffix in (".tif", ".tiff")) and filepath.stem[-3:].isdigit():
            # This looks like something123.tif.
            filepaths.append(filepath)
            continue
        outliers += 1

    fraction = outliers / (outliers + len(filepaths))
    if (outliers <= ABSOLUTE_THRESHOLD) and (fraction <= RELATIVE_THRESHOLD):
        # This looks like a TIFF sequence directory.
        import tifffile

        seq = tifffile.TiffSequence(sorted(filepaths))
        return TiffSequenceAdapter(seq)

    return None



import collections.abc
import os
import warnings

import h5py
import numpy

from ..adapters.utils import IndexersMixin
from ..iterviews import ItemsView, KeysView, ValuesView
from ..utils import DictView, node_repr
from .array import ArrayAdapter

SWMR_DEFAULT = bool(int(os.getenv("TILED_HDF5_SWMR_DEFAULT", "0")))


class HDF5DatasetAdapter(ArrayAdapter):
    def __init__(self, dataset):
        super().__init__(dataset, metadata=getattr(dataset, "attrs", {}))


class HDF5Adapter(collections.abc.Mapping, IndexersMixin):
    """
    Read an HDF5 file or a group within one.

    This map the structure of an HDF5 file onto a "Tree" of array structures.

    Examples
    --------

    From the root node of a file given a filepath

    >>> import h5py
    >>> HDF5Adapter.from_file("path/to/file.h5")

    From the root node of a file given an h5py.File object

    >>> import h5py
    >>> file = h5py.File("path/to/file.h5")
    >>> HDF5Adapter.from_file(file)

    From a group within a file

    >>> import h5py
    >>> file = h5py.File("path/to/file.h5")
    >>> HDF5Adapter(file["some_group']["some_sub_group"])

    """

    structure_family = "node"

    def __init__(self, node, *, specs=None, references=None, access_policy=None):
        if (access_policy is not None) and (
            not access_policy.check_compatibility(self)
        ):
            raise ValueError(
                f"Access policy {access_policy} is not compatible with this Tree."
            )
        self._node = node
        self._access_policy = access_policy
        self.specs = specs or []
        self.references = references or []
        super().__init__()

    @classmethod
    def from_file(
        cls, file, *, swmr=SWMR_DEFAULT, libver="latest", specs=None, references=None
    ):
        if not isinstance(file, h5py.File):
            file = h5py.File(file, "r", swmr=swmr, libver=libver)
        return cls(file, specs=specs, references=references)

    def __repr__(self):
        return node_repr(self, list(self))

    @property
    def access_policy(self):
        return self._access_policy

    @property
    def metadata(self):
        d = dict(self._node.attrs)
        for k, v in list(d.items()):
            # Convert any bytes to str.
            if isinstance(v, bytes):
                d[k] = v.decode()
        return DictView(d)

    def __iter__(self):
        yield from self._node

    def __getitem__(self, key):
        value = self._node[key]
        if isinstance(value, h5py.Group):
            return HDF5Adapter(value)
        else:
            if value.dtype == numpy.dtype("O"):
                warnings.warn(
                    f"The dataset {key} is of object type, using a "
                    "Python-only feature of h5py that is not supported by "
                    "HDF5 in general. Read more about that feature at "
                    "https://docs.h5py.org/en/stable/special.html. "
                    "Consider using a fixed-length field instead. "
                    "Tiled will serve an empty placeholder, unless the "
                    "object is of size 1, where it will attempt to repackage "
                    "the data into a numpy array."
                )

                check_str_dtype = h5py.check_string_dtype(value.dtype)
                if check_str_dtype.length is None:
                    dataset_names = value.file[self._node.name + "/" + key][...][()]
                    if value.size == 1:
                        arr = numpy.array(dataset_names)
                        return HDF5DatasetAdapter(arr)
                return HDF5DatasetAdapter(numpy.array([]))
            return HDF5DatasetAdapter(value)

    def __len__(self):
        return len(self._node)

    def keys(self):
        return KeysView(lambda: len(self), self._keys_slice)

    def values(self):
        return ValuesView(lambda: len(self), self._items_slice)

    def items(self):
        return ItemsView(lambda: len(self), self._items_slice)

    def search(self, query):
        """
        Return a Tree with a subset of the mapping.
        """
        raise NotImplementedError

    def read(self, fields=None):
        if fields is not None:
            raise NotImplementedError
        return self

    # The following two methods are used by keys(), values(), items().

    def _keys_slice(self, start, stop, direction):
        keys = list(self._node)
        if direction < 0:
            keys = reversed(keys)
        return keys[start:stop]

    def _items_slice(self, start, stop, direction):
        items = [(key, self[key]) for key in list(self)]
        if direction < 0:
            items = reversed(items)
        return items[start:stop]



import contextlib

import pytest

from ..client.utils import ClientError


def force_update(client):
    """
    Reach into the tree force it to process an updates. Block until complete.

    We could just wait (i.e. sleep) instead until the next polling loop completes,
    but this makes the tests take longer to run, and it can be flaky on CI services
    where things can take longer than they do in normal use.
    """
    client.context.http_client.app.state.root_tree.update_now()


@contextlib.contextmanager
def fail_with_status_code(status_code):
    with pytest.raises(ClientError) as info:
        yield
    assert info.value.response.status_code == status_code



from ..adapters.mapping import MapAdapter
from ..client import from_tree, record_history

tree = MapAdapter({})
client = from_tree(tree)


def test_history():
    "Very basic exercise of history"
    with record_history() as history:
        repr(client)  # trigger a request
    assert history.requests
    assert history.responses



import typer

profile_app = typer.Typer()


@profile_app.command("paths")
def profile_paths():
    "List the locations that the client will search for profiles (client-side configuration)."
    from ..profiles import paths

    print("\n".join(str(p) for p in paths))


@profile_app.command("list")
def profile_list():
    "List the profiles (client-side configuration) found and the files they were read from."
    from ..profiles import load_profiles

    profiles = load_profiles()
    if not profiles:
        typer.echo("No profiles found.")
        return
    max_len = max(len(name) for name in profiles)
    PADDING = 4

    print(
        "\n".join(
            f"{name:<{max_len + PADDING}}{filepath}"
            for name, (filepath, _) in profiles.items()
        )
    )


@profile_app.command("show")
def profile_show(profile_name: str):
    "Show the content of a profile."
    import sys

    import yaml

    from ..profiles import load_profiles

    profiles = load_profiles()
    try:
        filepath, content = profiles[profile_name]
    except KeyError:
        typer.echo(
            f"The profile {profile_name!r} could not be found. "
            "Use tiled profile list to see profile names.",
            err=True,
        )
        raise typer.Abort()
    print(f"Source: {filepath}", file=sys.stderr)
    print("--", file=sys.stderr)
    print(yaml.dump(content), file=sys.stdout)



import dask.base
import dask.dataframe
import pandas

from ..serialization.dataframe import serialize_arrow
from ..server.object_cache import NO_CACHE, get_object_cache
from ..structures.dataframe import DataFrameMacroStructure, DataFrameMicroStructure
from ..utils import DictView
from .array import ArrayAdapter


class DataFrameAdapter:
    """
    Wrap a dataframe-like object in an interface that Tiled can serve.

    Examples
    --------

    >>> df = pandas.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    >>> DataFrameAdapter(dask.dataframe.from_pandas(df), npartitions=1)

    Read a CSV (uses dask.dataframe.read_csv).

    >>> DataFrameAdapter.read_csv("myfiles.*.csv")
    >>> DataFrameAdapter.read_csv("s3://bucket/myfiles.*.csv")
    """

    structure_family = "dataframe"

    @classmethod
    def from_pandas(cls, *args, metadata=None, specs=None, references=None, **kwargs):
        ddf = dask.dataframe.from_pandas(*args, **kwargs)
        return cls.from_dask_dataframe(
            ddf, metadata=metadata, specs=specs, references=references
        )

    @classmethod
    def from_dask_dataframe(
        cls,
        ddf,
        metadata=None,
        specs=None,
        references=None,
    ):
        # Danger: using internal attribute _meta here.
        return cls(
            ddf.partitions,
            ddf._meta,
            ddf.divisions,
            metadata=metadata,
            specs=specs,
            references=references,
        )

    @classmethod
    def read_csv(cls, *args, metadata=None, specs=None, references=None, **kwargs):
        """
        Read a CSV.

        Internally, this uses dask.dataframe.read_csv.
        It forward all parameters to that function. See
        https://docs.dask.org/en/latest/dataframe-api.html#dask.dataframe.read_csv

        Examples
        --------

        >>> DataFrameAdapter.read_csv("myfiles.*.csv")
        >>> DataFrameAdapter.read_csv("s3://bucket/myfiles.*.csv")
        """
        ddf = dask.dataframe.read_csv(*args, **kwargs)
        # If an instance has previously been created using the same parameters,
        # then we are here because the caller wants a *fresh* view on this data.
        # Therefore, we should clear any cached data.
        cache = get_object_cache()
        if cache is not NO_CACHE:
            cache.discard_dask(ddf.__dask_keys__())
        return cls.from_dask_dataframe(
            ddf, metadata=metadata, specs=specs, references=references
        )

    read_csv.__doc__ = (
        """
    This wraps dask.dataframe.read_csv. Original docstring:

    """
        + dask.dataframe.read_csv.__doc__
    )

    def __init__(
        self, partitions, meta, divisions, *, metadata=None, specs=None, references=None
    ):
        self._metadata = metadata or {}
        self._partitions = list(partitions)
        self._meta = meta
        self._divisions = divisions
        self.specs = specs or []
        self.references = references or []

    def __repr__(self):
        return f"{type(self).__name__}({self._meta.columns!r})"

    def __getitem__(self, key):
        # Must compute to determine shape.
        return ArrayAdapter.from_array(self.read([key])[key].values)

    def items(self):
        yield from (
            (key, ArrayAdapter.from_array(self.read([key])[key].values))
            for key in self._meta.columns
        )

    @property
    def metadata(self):
        return DictView(self._metadata)

    def macrostructure(self):
        return DataFrameMacroStructure(
            columns=list(self._meta.columns), npartitions=len(self._partitions)
        )

    def microstructure(self):
        meta = bytes(serialize_arrow(self._meta, {}))
        divisions = bytes(
            serialize_arrow(pandas.DataFrame({"divisions": list(self._divisions)}), {})
        )
        return DataFrameMicroStructure(meta=meta, divisions=divisions)

    def read(self, fields=None):
        if any(p is None for p in self._partitions):
            raise ValueError("Not all partitions have been stored.")
        if isinstance(self._partitions[0], dask.dataframe.DataFrame):
            if fields is not None:
                ddf = dask.dataframe.concat(
                    [p[fields] for p in self._partitions], axis=0
                )
            else:
                ddf = dask.dataframe.concat(self._partitions, axis=0)
            # Note: If the cache is set to NO_CACHE, this is a null context.
            with get_object_cache().dask_context:
                return ddf.compute()
        df = pandas.concat(self._partitions, axis=0)
        if fields is not None:
            df = df[fields]
        return df

    def read_partition(self, partition, fields=None):
        partition = self._partitions[partition]
        if partition is None:
            raise RuntimeError(f"partition {partition} has not be stored yet")
        if fields is not None:
            partition = partition[fields]
        # Special case for dask to cache computed result in object cache.
        if isinstance(partition, dask.dataframe.DataFrame):
            # Note: If the cache is set to NO_CACHE, this is a null context.
            with get_object_cache().dask_context:
                return partition.compute()
        return partition



from typing import Optional

try:
    import typer
except Exception as err:
    raise Exception(
        """
from ._admin import admin_app
from ._api_key import api_key_app
from ._profile import profile_app
from ._serve import serve_app

You are trying to the run the tiled commandline tool but you do not have the
necessary dependencies. It looks like tiled has been installed with
bare-minimum dependencies, possibly via

    pip install tiled

Instead, try:

    pip install tiled[all]  # Note: on a Mac, you may need quotes like 'tiled[all]'.

which installs *everything* you might want. For other options, see:

    https://blueskyproject.io/tiled/tutorials/installation.html
"""
    ) from err

cli_app = typer.Typer()

from ._admin import admin_app  # noqa: E402
from ._api_key import api_key_app  # noqa: E402
from ._profile import profile_app  # noqa: E402
from ._serve import serve_app  # noqa: E402
from ._utils import (  # noqa E402
    CLI_CACHE_DIR,
    get_context,
    get_default_profile_name,
    get_profile,
)

cli_app.add_typer(serve_app, name="serve", help="Launch a Tiled server.")
cli_app.add_typer(
    profile_app, name="profile", help="Examine Tiled 'profiles' (client-side config)."
)
cli_app.add_typer(
    api_key_app, name="api_key", help="Create, list, and revoke API keys."
)
cli_app.add_typer(
    admin_app,
    name="admin",
    help="Administrative utilities for managing large deployments.",
)


@cli_app.command("connect")
def connect(
    uri_or_profile: str = typer.Argument(
        ..., help="URI 'http[s]://...' or a profile name."
    ),
    no_verify: bool = typer.Option(False, "--no-verify", help="Skip SSL verification."),
):
    """
    "Connect" to a Tiled server; set it as default.
    """
    from ..client.context import Context
    from ..profiles import list_profiles, load_profiles, paths

    user_profiles_dir = paths[-1]
    if uri_or_profile.startswith("http://") or uri_or_profile.startswith("https://"):
        # This looks like a URI.
        uri = uri_or_profile
        name = "auto"
        Context.from_any_uri(uri, verify=not no_verify)
        user_profiles_dir.mkdir(parents=True, exist_ok=True)
        with open(user_profiles_dir / "auto.yml", "w") as file:
            file.write(
                f"""# This file is managed by the Tiled CLI.
# Any edits made by hand may be discarded.
auto:
  uri: {uri}
  verify: {"true" if not no_verify else "false"}
"""
            )
    else:
        # Is this a profile name?
        if uri_or_profile in list_profiles():
            name = uri_or_profile
            _, profile = load_profiles()[name]
            if "direct" in profile:
                raise ValueError(
                    f"Profile {profile} uses in a direct (in-process) Tiled server "
                    "and cannot be connected to from the CLI."
                )
            options = {"verify": profile.get("verify", True)}
            if no_verify:
                options["verify"] = False
            Context.from_any_uri(profile["uri"], **options)
        else:
            typer.echo(
                f"Not sure what to do with tree {uri_or_profile!r}. "
                "It does not look like a URI (it does not start with http[s]://) "
                "and it does not match any profiles. Use `tiled profiles list` to "
                "see profiles.",
                err=True,
            )
            raise typer.Abort()
    CLI_CACHE_DIR.mkdir(parents=True, exist_ok=True)
    with open(CLI_CACHE_DIR / "active_profile", "w") as file:
        file.write(name)
    typer.echo(f"Tiled profile {name!r} is set as the default.")


@cli_app.command("status")
def status():
    """
    Show the current default Tiled server.
    """
    name = get_default_profile_name()
    if name is None:
        typer.echo("Not connected.", err=True)
    else:
        typer.echo(f"Using profile {name!r}\n", err=True)
        import yaml

        _, profile_content = get_profile(name)
        typer.echo(yaml.dump(profile_content))


@cli_app.command("disconnect")
def disconnect():
    """
    "Disconnect" from the default Tiled server.
    """
    filepath = CLI_CACHE_DIR / "active_profile"
    # filepath.unlink(missing_ok=False)  # Python 3.8+
    try:
        filepath.unlink()
    except FileNotFoundError:
        pass


@cli_app.command("login")
def login(
    profile: Optional[str] = typer.Option(
        None, help="If you use more than one Tiled server, use this to specify which."
    ),
    show_secret_tokens: bool = typer.Option(
        False, "--show-secret-tokens", help="Show secret tokens after successful login."
    ),
):
    """
    Log in to an authenticated Tiled server.
    """
    import json

    from ..client.context import Context

    profile_name, profile_content = get_profile(profile)
    options = {"verify": profile_content.get("verify", True)}
    context, _ = Context.from_any_uri(profile_content["uri"], **options)
    provider_spec, username = context.authenticate()
    filepath = CLI_CACHE_DIR / "profile_auths"
    filepath.mkdir(parents=True, exist_ok=True)
    with open(filepath / profile_name, "w") as file:
        json.dump({"provider": provider_spec["provider"], "username": username}, file)
    if show_secret_tokens:
        typer.echo(json.dumps(dict(context.tokens), indent=4))


@cli_app.command("logout")
def logout(
    profile: Optional[str] = typer.Option(
        None, help="If you use more than one Tiled server, use this to specify which."
    ),
):
    """
    Log out from one or all authenticated Tiled servers.
    """
    import json

    from ..client.context import Context

    profile_name, profile_content = get_profile(profile)
    filepath = CLI_CACHE_DIR / "profile_auths" / profile_name
    context, _ = Context.from_any_uri(
        profile_content["uri"], verify=profile_content.get("verify", True)
    )
    if filepath.is_file():
        with open(filepath, "r") as file:
            auth = json.load(file)
        context.authenticate(auth["username"], auth["provider"])
    context.logout()
    # filepath.unlink(missing_ok=False)  # Python 3.8+
    try:
        filepath.unlink()
    except FileNotFoundError:
        pass


@cli_app.command("tree")
def tree(
    profile: Optional[str] = typer.Option(
        None, help="If you use more than one Tiled server, use this to specify which."
    ),
    max_lines: int = typer.Argument(20, help="Max lines to show."),
):
    """
    Show the names of entries in a Tree.

    This is similar to the UNIX utility `tree` for listing nested directories.
    """
    from ..client.constructors import from_context
    from ..utils import gen_tree

    context = get_context(profile)
    client = from_context(context)
    for counter, line in enumerate(gen_tree(client), start=1):
        if (max_lines is not None) and (counter > max_lines):
            print(
                f"Output truncated at {max_lines} lines. "
                "Use `tiled tree <N>` to see <N> lines."
            )
            break
        print(line)


@cli_app.command("download")
def download(
    profile: Optional[str] = typer.Option(
        None, help="If you use more than one Tiled server, use this to specify which."
    ),
    cache_path: str = typer.Argument(..., help="Local directory for cache storage"),
    capacity: Optional[int] = typer.Argument(None, help="Max capacity in bytes"),
):
    """
    Download content from a Tree to an on-disk cache.
    """
    from ..client.cache import Cache, download
    from ..client.constructors import from_profile

    profile_name, _ = get_profile(profile)

    cache = Cache.on_disk(cache_path, capacity=capacity)
    client = from_profile(profile_name, cache=cache)
    download(client)


main = cli_app


if __name__ == "__main__":
    main()

# This object is used by the auto-generated documentation.
typer_click_object = typer.main.get_command(cli_app)



import httpx

from ..adapters.mapping import MapAdapter
from ..client import from_tree

tree = MapAdapter({})


def test_configurable_timeout():
    c = from_tree(tree)
    assert c.context.http_client.timeout.connect != 17
    c = from_tree(tree, timeout=httpx.Timeout(17))
    assert c.context.http_client.timeout.connect == 17



"""
This tests tiled's validation registry
"""

import numpy as np
import pandas as pd

from ..client import from_config, from_tree
from ..validation_registration import ValidationError, ValidationRegistry
from .utils import fail_with_status_code
from .writable_adapters import WritableMapAdapter

API_KEY = "secret"


def lower_case_dict(d):
    out = {}
    modified = False

    for k, v in d.items():
        if isinstance(k, str) and not k.islower():
            out[k.lower()] = v
            modified = True
        else:
            out[k] = v

    return out, modified


def validate_foo(metadata, structure_family, structure, spec, references):
    if structure_family != "dataframe":
        raise ValidationError(f"structure family for spec {spec} must be dataframe")

    if list(structure.macro.columns) != ["a", "b"]:
        raise ValidationError(f"structure for spec {spec} must have columns ['a', 'b']")

    metadata, metadata_modified = lower_case_dict(metadata)

    if "foo" not in metadata:
        raise ValidationError("metadata for spec {spec} must contain foo")

    if metadata_modified:
        return metadata


def test_validators():

    validation_registry = ValidationRegistry()
    validation_registry.register("foo", validate_foo)

    tree = WritableMapAdapter({})
    client = from_tree(
        tree,
        api_key=API_KEY,
        authentication={"single_user_api_key": API_KEY},
        validation_registry=validation_registry,
    )

    with fail_with_status_code(400):
        a = np.ones((5, 7))
        client.write_array(a, metadata={}, specs=["foo"])

    with fail_with_status_code(400):
        df = pd.DataFrame({"x": np.zeros(100), "y": np.zeros(100)})
        client.write_dataframe(df, metadata={}, specs=["foo"])

    with fail_with_status_code(400):
        df = pd.DataFrame({"a": np.zeros(100), "b": np.zeros(100)})
        client.write_dataframe(df, metadata={}, specs=["foo"])

    metadata = {"id": 1, "foo": "bar"}
    df = pd.DataFrame({"a": np.zeros(100), "b": np.zeros(100)})
    result = client.write_dataframe(df, metadata=metadata, specs=["foo"])
    assert result.metadata == metadata
    result_df = result.read()
    pd.testing.assert_frame_equal(result_df, df)

    metadata_upper = {"ID": 2, "FOO": "bar"}
    metadata_lower, _ = lower_case_dict(metadata_upper)
    result = client.write_dataframe(df, metadata=metadata_upper, specs=["foo"])
    assert result.metadata == metadata_lower
    result_df = result.read()
    pd.testing.assert_frame_equal(result_df, df)


tree = WritableMapAdapter({})


def test_unknown_spec():
    "Test unknown spec rejected for upload."
    config = {
        "trees": [{"tree": f"{__name__}:tree", "path": "/"}],
        "specs": [{"spec": "a"}],
        "authentication": {"single_user_api_key": API_KEY},
    }
    client = from_config(config, api_key=API_KEY)
    a = np.ones((5, 7))
    client.write_array(a, metadata={}, specs=["a"])

    with fail_with_status_code(400):
        client.write_array(a, metadata={}, specs=["b"])
